import contextlib
import json
import os
import shutil
import tempfile
from unittest import mock
from urllib.parse import quote as url_quote

import requests

from pylxd import exceptions, models
from pylxd.tests import testing


class TestInstance(testing.PyLXDTestCase):
    """Tests for pylxd.models.Instance."""

    def test_all(self):
        """A list of all instances are returned."""
        instances = models.Instance.all(self.client)

        self.assertEqual(1, len(instances))

    def test_get(self):
        """Return a instance."""
        name = "an-instance"

        an_instance = models.Instance.get(self.client, name)

        self.assertEqual(name, an_instance.name)

    def test_get_not_found(self):
        """LXDAPIException is raised when the instance doesn't exist."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/instances/an-missing-instance$",
            }
        )

        name = "an-missing-instance"

        self.assertRaises(
            exceptions.LXDAPIException, models.Instance.get, self.client, name
        )

    def test_get_error(self):
        """LXDAPIException is raised when the LXD API errors."""

        def not_found(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/instances/an-missing-instance$",
            }
        )

        name = "an-missing-instance"

        self.assertRaises(
            exceptions.LXDAPIException, models.Instance.get, self.client, name
        )

    def test_create(self):
        """A new instance is created."""
        config = {"name": "an-new-instance"}

        an_new_instance = models.Instance.create(self.client, config, wait=True)

        self.assertEqual(config["name"], an_new_instance.name)

    def test_create_remote_location(self):
        """A new instance is created at target."""
        config = {"name": "an-new-remote-instance"}

        # the server must be in a cluster for the location to be set
        self.client.host_info["environment"]["server_clustered"] = True

        an_new_remote_instance = models.Instance.create(
            self.client, config, wait=True, target="an-remote"
        )

        self.assertEqual(config["name"], an_new_remote_instance.name)
        self.assertEqual("an-remote", an_new_remote_instance.location)

    def test_create_location_none(self):
        config = {"name": "an-new-remote-instance"}

        instance = models.Instance.create(self.client, config, wait=True)

        self.assertIsNone(instance.location)

    def test_exists(self):
        """A instance exists."""
        name = "an-instance"

        self.assertTrue(models.Instance.exists(self.client, name))

    def test_not_exists(self):
        """A instance exists."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/instances/an-missing-instance$",
            }
        )

        name = "an-missing-instance"

        self.assertFalse(models.Instance.exists(self.client, name))

    def test_fetch(self):
        """A sync updates the properties of a instance."""
        an_instance = models.Instance(self.client, name="an-instance")

        an_instance.sync()

        self.assertTrue(an_instance.ephemeral)

    def test_fetch_not_found(self):
        """LXDAPIException is raised on a 404 for updating instance."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/instances/an-missing-instance$",
            }
        )

        an_instance = models.Instance(self.client, name="an-missing-instance")

        self.assertRaises(exceptions.LXDAPIException, an_instance.sync)

    def test_fetch_error(self):
        """LXDAPIException is raised on error."""

        def not_found(request, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "An bad error", "error_code": 500}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/instances/an-missing-instance$",
            }
        )

        an_instance = models.Instance(self.client, name="an-missing-instance")

        self.assertRaises(exceptions.LXDAPIException, an_instance.sync)

    def test_update(self):
        """A instance is updated."""
        an_instance = models.Instance(self.client, name="an-instance")
        an_instance.architecture = 1
        an_instance.config = {}
        an_instance.created_at = 1
        an_instance.devices = {}
        an_instance.ephemeral = 1
        an_instance.expanded_config = {}
        an_instance.expanded_devices = {}
        an_instance.profiles = 1
        an_instance.status = 1

        an_instance.save(wait=True)

        self.assertTrue(an_instance.ephemeral)

    def test_rename(self):
        an_instance = models.Instance(self.client, name="an-instance")

        an_instance.rename("an-renamed-instance", wait=True)

        self.assertEqual("an-renamed-instance", an_instance.name)

    def test_delete(self):
        """A instance is deleted."""
        # XXX: rockstar (21 May 2016) - This just executes
        # a code path. There should be an assertion here, but
        # it's not clear how to assert that, just yet.
        an_instance = models.Instance(self.client, name="an-instance")

        an_instance.delete(wait=True)

    @mock.patch("pylxd.models.instance._StdinWebsocket")
    @mock.patch("pylxd.models.instance._CommandWebsocketClient")
    def test_execute(self, _CommandWebsocketClient, _StdinWebsocket):
        """A command is executed on a instance."""
        fake_websocket = mock.Mock()
        fake_websocket.data = "test\n"
        _StdinWebsocket.return_value = fake_websocket
        _CommandWebsocketClient.return_value = fake_websocket

        an_instance = models.Instance(self.client, name="an-instance")

        result = an_instance.execute(["echo", "test"])

        self.assertEqual(0, result.exit_code)
        self.assertEqual("test\n", result.stdout)

    @mock.patch("pylxd.models.instance._StdinWebsocket")
    @mock.patch("pylxd.models.instance._CommandWebsocketClient")
    def test_execute_uses_client_ssl_options(
        self, _CommandWebsocketClient, _StdinWebsocket
    ):
        """Websockets are created using the client's ssl_options."""
        fake_websocket = mock.Mock()
        fake_websocket.data = "test\n"
        self.client.ssl_options = {"certfile": "/foo", "keyfile": "/bar"}
        _StdinWebsocket.return_value = fake_websocket
        _CommandWebsocketClient.return_value = fake_websocket

        an_instance = models.Instance(self.client, name="an-instance")

        result = an_instance.execute(["echo", "test"])

        self.assertEqual(
            _StdinWebsocket.call_args.kwargs["ssl_options"],
            {"certfile": "/foo", "keyfile": "/bar"},
        )
        self.assertEqual(
            _CommandWebsocketClient.call_args.kwargs["ssl_options"],
            {"certfile": "/foo", "keyfile": "/bar"},
        )

        self.assertEqual(0, result.exit_code)
        self.assertEqual("test\n", result.stdout)

    @mock.patch("pylxd.models.instance._StdinWebsocket")
    @mock.patch("pylxd.models.instance._CommandWebsocketClient")
    def test_execute_with_env(self, _CommandWebsocketClient, _StdinWebsocket):
        """A command is executed on a instance with custom env variables."""
        fake_websocket = mock.Mock()
        fake_websocket.data = "test\n"
        _StdinWebsocket.return_value = fake_websocket
        _CommandWebsocketClient.return_value = fake_websocket

        an_instance = models.Instance(self.client, name="an-instance")

        result = an_instance.execute(["echo", "test"], environment={"DISPLAY": ":1"})

        self.assertEqual(0, result.exit_code)
        self.assertEqual("test\n", result.stdout)

    def test_execute_string(self):
        """A command passed as string raises a TypeError."""
        an_instance = models.Instance(self.client, name="an-instance")

        self.assertRaises(TypeError, an_instance.execute, "apt-get update")

    def test_raw_interactive_execute(self):
        an_instance = models.Instance(self.client, name="an-instance")

        result = an_instance.raw_interactive_execute(["/bin/bash"])

        self.assertEqual(
            result["ws"], "/1.0/operations/operation-abc/websocket?secret=abc"
        )
        self.assertEqual(
            result["control"], "/1.0/operations/operation-abc/websocket?secret=jkl"
        )

    def test_raw_interactive_execute_env(self):
        an_instance = models.Instance(self.client, name="an-instance")

        result = an_instance.raw_interactive_execute(["/bin/bash"], {"PATH": "/"})

        self.assertEqual(
            result["ws"], "/1.0/operations/operation-abc/websocket?secret=abc"
        )
        self.assertEqual(
            result["control"], "/1.0/operations/operation-abc/websocket?secret=jkl"
        )

    def test_raw_interactive_execute_string(self):
        """A command passed as string raises a TypeError."""
        an_instance = models.Instance(self.client, name="an-instance")

        self.assertRaises(
            TypeError, an_instance.raw_interactive_execute, "apt-get update"
        )

    def test_raw_interactive_execute_options(self):
        """It's possible to pass user, group and cwd arguments to an execute command."""
        an_instance = models.Instance(self.client, name="an-instance")

        result = an_instance.raw_interactive_execute(
            ["/bin/bash"], user="user", group="group", cwd="/some/path"
        )
        self.assertEqual(
            result["ws"], "/1.0/operations/operation-abc/websocket?secret=abc"
        )
        self.assertEqual(
            result["control"], "/1.0/operations/operation-abc/websocket?secret=jkl"
        )

    def test_migrate(self):
        """A instance is migrated."""
        from pylxd.client import Client

        client2 = Client(endpoint="http://pylxd2.test")
        an_instance = models.Instance(self.client, name="an-instance")

        an_migrated_instance = an_instance.migrate(client2)

        self.assertEqual("an-instance", an_migrated_instance.name)
        self.assertEqual(client2, an_migrated_instance.client)

    @mock.patch("pylxd.models.instance.Instance.generate_migration_data")
    def test_migrate_exception_error(self, generate_migration_data):
        """LXDAPIException is raised in case of migration failure"""
        from pylxd.client import Client
        from pylxd.exceptions import LXDAPIException

        def generate_exception(*args, **kwargs):
            response = mock.Mock()
            response.status_code = 400
            raise LXDAPIException(response)

        generate_migration_data.side_effect = generate_exception

        an_instance = models.Instance(self.client, name="an-instance")

        client2 = Client(endpoint="http://pylxd2.test")
        self.assertRaises(LXDAPIException, an_instance.migrate, client2)

    @mock.patch("pylxd.models.instance.Instance.generate_migration_data")
    def test_migrate_exception_running(self, generate_migration_data):
        """Migrated instance already running on destination"""
        from pylxd.client import Client
        from pylxd.exceptions import LXDAPIException

        client2 = Client(endpoint="http://pylxd2.test")
        an_instance = models.Instance(self.client, name="an-instance")
        an_instance.status_code = 103

        def generate_exception(*args, **kwargs):
            response = mock.Mock()
            response.status_code = 103
            raise LXDAPIException(response)

        generate_migration_data.side_effect = generate_exception

        an_migrated_instance = an_instance.migrate(client2, live=True)

        self.assertEqual("an-instance", an_migrated_instance.name)
        self.assertEqual(client2, an_migrated_instance.client)
        generate_migration_data.assert_called_once_with(True)

    def test_migrate_started(self):
        """A instance is migrated."""
        from pylxd.client import Client

        client2 = Client(endpoint="http://pylxd2.test")
        an_instance = models.Instance.get(self.client, name="an-instance")
        an_instance.status_code = 103

        an_migrated_instance = an_instance.migrate(client2)

        self.assertEqual("an-instance", an_migrated_instance.name)
        self.assertEqual(client2, an_migrated_instance.client)

    def test_migrate_stopped(self):
        """A instance is migrated."""
        from pylxd.client import Client

        client2 = Client(endpoint="http://pylxd2.test")
        an_instance = models.Instance.get(self.client, name="an-instance")
        an_instance.status_code = 102

        an_migrated_instance = an_instance.migrate(client2)

        self.assertEqual("an-instance", an_migrated_instance.name)
        self.assertEqual(client2, an_migrated_instance.client)

    @mock.patch("pylxd.client._APINode.get")
    def test_migrate_local_client(self, get):
        """Migration from local clients is not supported."""
        # Mock out the _APINode for the local instance.
        response = mock.Mock()
        response.json.return_value = {"metadata": {"fake": "response"}}
        response.status_code = 200
        get.return_value = response

        from pylxd.client import Client

        client2 = Client(endpoint="http+unix://pylxd2.test")
        an_instance = models.Instance(client2, name="an-instance")

        self.assertRaises(ValueError, an_instance.migrate, self.client)

    def test_publish(self):
        """Instances can be published."""
        self.add_rule(
            {
                "text": json.dumps(
                    {
                        "type": "sync",
                        "metadata": {
                            "id": "operation-abc",
                            "metadata": {
                                "fingerprint": (
                                    "e3b0c44298fc1c149afbf4c8996fb92427"
                                    "ae41e4649b934ca495991b7852b855"
                                )
                            },
                        },
                    }
                ),
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/operations/operation-abc$",
            }
        )

        an_instance = models.Instance(self.client, name="an-instance")
        # Hack to get around mocked data
        an_instance.type = "container"

        image = an_instance.publish(wait=True)

        self.assertEqual(
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            image.fingerprint,
        )

    @mock.patch("pylxd.client._APINode.put")
    def test_restore_snapshot(self, put):
        """Snapshots can be restored"""
        response = mock.MagicMock(status_code=202)
        put.return_value = response

        an_instance = models.Instance(self.client, name="an-instance")
        an_instance.restore_snapshot("thing")

        put.assert_called_once_with(json={"restore": "thing", "stateful": False})

    @mock.patch("pylxd.client._APINode.put")
    def test_restore_snapshot_stateful(self, put):
        """Snapshots can be restored"""
        response = mock.Mock()
        response.json.return_value = {}
        response.status_code = 202
        put.return_value = response

        an_instance = models.Instance(self.client, name="an-instance")
        an_instance.restore_snapshot("thing", stateful=True)

        put.assert_called_once_with(json={"restore": "thing", "stateful": True})


class TestInstanceState(testing.PyLXDTestCase):
    """Tests for pylxd.models.InstanceState."""

    def test_get(self):
        """Return a instance."""
        name = "an-instance"

        an_instance = models.Instance.get(self.client, name)
        state = an_instance.state()

        self.assertEqual("Running", state.status)
        self.assertEqual(103, state.status_code)

    def test_start(self):
        """A instance is started."""
        an_instance = models.Instance.get(self.client, "an-instance")

        an_instance.start(wait=True)

    def test_stop(self):
        """A instance is stopped."""
        an_instance = models.Instance.get(self.client, "an-instance")

        an_instance.stop()

    def test_restart(self):
        """A instance is restarted."""
        an_instance = models.Instance.get(self.client, "an-instance")

        an_instance.restart()

    def test_freeze(self):
        """A instance is suspended."""
        an_instance = models.Instance.get(self.client, "an-instance")

        an_instance.freeze()

    def test_unfreeze(self):
        """A instance is resumed."""
        an_instance = models.Instance.get(self.client, "an-instance")

        an_instance.unfreeze()


class TestInstanceSnapshots(testing.PyLXDTestCase):
    """Tests for pylxd.models.Instance.snapshots."""

    def setUp(self):
        super().setUp()
        self.instance = models.Instance.get(self.client, "an-instance")

    def test_get(self):
        """Return a specific snapshot."""
        snapshot = self.instance.snapshots.get("an-snapshot")

        self.assertEqual("an-snapshot", snapshot.name)

    def test_all(self):
        """Return all snapshots."""
        snapshots = self.instance.snapshots.all()

        self.assertEqual(1, len(snapshots))
        self.assertEqual("an-snapshot", snapshots[0].name)
        self.assertEqual(self.client, snapshots[0].client)
        self.assertEqual(self.instance, snapshots[0].instance)

    def test_create(self):
        """Create a snapshot."""
        snapshot = self.instance.snapshots.create(
            "an-snapshot", stateful=True, wait=True
        )

        self.assertEqual("an-snapshot", snapshot.name)


class TestSnapshot(testing.PyLXDTestCase):
    """Tests for pylxd.models.Snapshot."""

    def setUp(self):
        super().setUp()
        self.instance = models.Instance.get(self.client, "an-instance")

    def test_rename(self):
        """A snapshot is renamed."""
        snapshot = models.Snapshot(
            self.client, instance=self.instance, name="an-snapshot"
        )

        snapshot.rename("an-renamed-snapshot", wait=True)

        self.assertEqual("an-renamed-snapshot", snapshot.name)

    def test_delete(self):
        """A snapshot is deleted."""
        snapshot = models.Snapshot(
            self.client, instance=self.instance, name="an-snapshot"
        )

        snapshot.delete(wait=True)

        # TODO: add an assertion here

    def test_delete_failure(self):
        """If the response indicates delete failure, raise an exception."""

        def not_found(request, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "DELETE",
                "url": (
                    r"^http://pylxd.test/1.0/instances/"
                    "an-instance/snapshots/an-snapshot$"
                ),
            }
        )

        snapshot = models.Snapshot(
            self.client, instance=self.instance, name="an-snapshot"
        )

        self.assertRaises(exceptions.LXDAPIException, snapshot.delete)

    def test_publish(self):
        """Snapshots can be published."""
        self.add_rule(
            {
                "text": json.dumps(
                    {
                        "type": "sync",
                        "metadata": {
                            "id": "operation-abc",
                            "metadata": {
                                "fingerprint": (
                                    "e3b0c44298fc1c149afbf4c8996fb92427"
                                    "ae41e4649b934ca495991b7852b855"
                                )
                            },
                        },
                    }
                ),
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/operations/operation-abc$",
            }
        )

        snapshot = models.Snapshot(
            self.client, instance=self.instance, name="an-snapshot"
        )

        image = snapshot.publish(wait=True)

        self.assertEqual(
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            image.fingerprint,
        )

    @mock.patch("pylxd.models.Instance.restore_snapshot")
    def test_restore_snapshot(self, restore_snapshot):
        """Snapshots can be restored from the snapshot object"""
        response = mock.MagicMock(status_code=202)
        restore_snapshot.return_value = response

        snapshot = models.Snapshot(
            self.client, instance=self.instance, name="an-snapshot"
        )
        snapshot.restore(wait=True)
        restore_snapshot.assert_called_once_with("an-snapshot", True, stateful=False)

    @mock.patch("pylxd.models.Instance.restore_snapshot")
    def test_restore_snapshot_stateful(self, restore_snapshot):
        """Stateful Snapshots can be restored from the snapshot object"""
        response = mock.MagicMock(status_code=202)
        restore_snapshot.return_value = response

        snapshot = models.Snapshot(
            self.client, instance=self.instance, name="an-snapshot", stateful=True
        )
        snapshot.restore(wait=False)
        restore_snapshot.assert_called_once_with("an-snapshot", False, stateful=True)


class TestFiles(testing.PyLXDTestCase):
    """Tests for pylxd.models.Instance.files."""

    def setUp(self):
        super().setUp()
        self.instance = models.Instance.get(self.client, "an-instance")

    def test_put_delete(self):
        """A file is put on the instance and then deleted"""
        # we are mocked, so delete should initially not be available
        self.assertEqual(False, self.instance.files.delete_available())
        self.assertRaises(
            exceptions.LXDAPIExtensionNotAvailable,
            self.instance.files.delete,
            "/some/file",
        )
        # Now insert delete
        self.add_rule(
            {
                "text": json.dumps(
                    {
                        "type": "sync",
                        "metadata": {
                            "auth": "trusted",
                            "environment": {
                                "certificate": "an-pem-cert",
                            },
                            "api_extensions": ["file_delete"],
                        },
                    }
                ),
                "method": "GET",
                "url": r"^http://pylxd.test/1.0$",
            }
        )

        # Update hostinfo
        self.client.host_info = self.client.api.get().json()["metadata"]

        self.assertEqual(True, self.instance.files.delete_available())

        # mock out the delete rule:
        self.add_rule(
            {
                "method": "DELETE",
                "url": (
                    r"^http://pylxd.test/1.0/instances/an-instance/files"
                    r"\?path=%2Fsome%2Ffile$"
                ),
            }
        )
        self.instance.files.delete("/some/file")

        # now check that an error (non 200) causes an exception
        def responder(request, context):
            context.status_code = 404

        self.add_rule(
            {
                "text": responder,
                "method": "DELETE",
                "url": (
                    r"^http://pylxd.test/1.0/instances/an-instance/files"
                    r"\?path=%2Fsome%2Ffile%2Fnot%2Ffound$"
                ),
            }
        )
        with self.assertRaises(exceptions.LXDAPIException):
            self.instance.files.delete("/some/file/not/found")

    def test_put_mode_uid_gid(self):
        """Should be able to set the mode, uid and gid of a file"""
        # fix up the default POST rule to allow us to see the posted vars
        _capture = {}

        def capture(request, context):
            _capture["headers"] = getattr(request._request, "headers")
            context.status_code = 200

        self.add_rule(
            {
                "text": capture,
                "method": "POST",
                "url": (
                    r"^http://pylxd.test/1.0/instances/an-instance/files"
                    r"\?path=%2Ftmp%2Fputted$"
                ),
            }
        )

        data = "The quick brown fox"
        # start with an octal mode
        self.instance.files.put("/tmp/putted", data, mode=0o123, uid=1, gid=2)
        headers = _capture["headers"]
        self.assertEqual(headers["X-LXD-mode"], "0123")
        self.assertEqual(headers["X-LXD-uid"], "1")
        self.assertEqual(headers["X-LXD-gid"], "2")
        # use a str mode this type
        self.instance.files.put("/tmp/putted", data, mode="456")
        headers = _capture["headers"]
        self.assertEqual(headers["X-LXD-mode"], "0456")
        # check that mode='0644' also works (i.e. already has 0 prefix)
        self.instance.files.put("/tmp/putted", data, mode="0644")
        headers = _capture["headers"]
        self.assertEqual(headers["X-LXD-mode"], "0644")
        # check that assertion is raised
        with self.assertRaises(ValueError):
            self.instance.files.put("/tmp/putted", data, mode=object)

    def test_mk_dir(self):
        """Tests pushing an empty directory"""
        _capture = {}

        def capture(request, context):
            _capture["headers"] = getattr(request._request, "headers")
            context.status_code = 200

        self.add_rule(
            {
                "text": capture,
                "method": "POST",
                "url": (
                    r"^http://pylxd.test/1.0/instances/an-instance/files"
                    r"\?path=%2Ftmp%2Fputted$"
                ),
            }
        )

        self.instance.files.mk_dir("/tmp/putted", mode=0o123, uid=1, gid=2)
        headers = _capture["headers"]
        self.assertEqual(headers["X-LXD-type"], "directory")
        self.assertEqual(headers["X-LXD-mode"], "0123")
        self.assertEqual(headers["X-LXD-uid"], "1")
        self.assertEqual(headers["X-LXD-gid"], "2")
        # check that assertion is raised
        with self.assertRaises(ValueError):
            self.instance.files.mk_dir("/tmp/putted", mode=object)

        response = mock.Mock()
        response.status_code = 404

        with mock.patch("pylxd.client._APINode.post", response):
            with self.assertRaises(exceptions.LXDAPIException):
                self.instance.files.mk_dir("/tmp/putted")

    def test_recursive_put(self):
        @contextlib.contextmanager
        def tempdir(prefix="tmp"):
            tmpdir = tempfile.mkdtemp(prefix=prefix)
            try:
                yield tmpdir
            finally:
                shutil.rmtree(tmpdir)

        def create_file(_dir, name, content):
            path = os.path.join(_dir, name)
            actual_dir = os.path.dirname(path)
            if not os.path.exists(actual_dir):
                os.makedirs(actual_dir)
            with open(path, "w") as f:
                f.write(content)

        _captures = []

        def capture(request, context):
            _captures.append(
                {
                    "headers": getattr(request._request, "headers"),
                    "body": request._request.body,
                }
            )
            context.status_code = 200

        with tempdir() as _dir:
            base = r"^http://pylxd.test/1.0/instances/" r"an-instance/files\?path="
            rules = [
                {
                    "text": capture,
                    "method": "POST",
                    "url": base + url_quote("target", safe="") + "$",
                },
                {
                    "text": capture,
                    "method": "POST",
                    "url": base + url_quote("target/dir", safe="") + "$",
                },
                {
                    "text": capture,
                    "method": "POST",
                    "url": base + url_quote("target/file1", safe="") + "$",
                },
                {
                    "text": capture,
                    "method": "POST",
                    "url": base + url_quote("target/dir/file2", safe="") + "$",
                },
            ]
            self.add_rules(rules)

            create_file(_dir, "file1", "This is file1")
            create_file(_dir, "dir/file2", "This is file2")

            self.instance.files.recursive_put(_dir, "./target/")

            self.assertEqual(_captures[0]["headers"]["X-LXD-type"], "directory")
            self.assertEqual(_captures[1]["body"], b"This is file1")
            self.assertEqual(_captures[2]["headers"]["X-LXD-type"], "directory")
            self.assertEqual(_captures[3]["body"], b"This is file2")

    def test_get(self):
        """A file is retrieved from the instance."""
        data = self.instance.files.get("/tmp/getted")

        self.assertEqual(b"This is a getted file", data)

    def test_recursive_get(self):
        """A folder is retrieved recursively from the instance"""
        response = requests.models.Response()
        response.status_code = 200
        response.headers["X-LXD-type"] = "directory"
        response._content = json.dumps({"metadata": ["file1", "file2"]})

        response1 = requests.models.Response()
        response1.status_code = 200
        response1.headers["X-LXD-type"] = "file"
        response1._content = "This is file1"

        response2 = requests.models.Response()
        response2.status_code = 200
        response2.headers["X-LXD-type"] = "file"
        response2._content = "This is file2"

        return_values = [response, response1, response2]

        with mock.patch("pylxd.client._APINode.get") as get_mocked:
            get_mocked.side_effect = return_values
            with mock.patch("os.mkdir") as mkdir_mocked:
                mock_open = mock.mock_open()
                with mock.patch("pylxd.models.instance.open", mock_open):
                    self.instance.files.recursive_get("/tmp/getted", "/tmp")
                    assert mkdir_mocked.call_count == 1
                    assert mock_open.call_count == 2

    def test_get_not_found(self):
        """LXDAPIException is raised on bogus filenames."""

        def not_found(request, context):
            context.status_code = 500

        rule = {
            "text": not_found,
            "method": "GET",
            "url": (
                r"^http://pylxd.test/1.0/instances/an-instance/files"
                r"\?path=%2Ftmp%2Fgetted$"
            ),
        }
        self.add_rule(rule)

        self.assertRaises(
            exceptions.LXDAPIException, self.instance.files.get, "/tmp/getted"
        )

    def test_get_error(self):
        """LXDAPIException is raised on error."""

        def not_found(request, context):
            context.status_code = 503

        rule = {
            "text": not_found,
            "method": "GET",
            "url": (
                r"^http://pylxd.test/1.0/instances/an-instance/files"
                r"\?path=%2Ftmp%2Fgetted$"
            ),
        }
        self.add_rule(rule)

        self.assertRaises(
            exceptions.LXDAPIException, self.instance.files.get, "/tmp/getted"
        )

    # for bug/281 -- getting an empty json file is interpreted as an API
    # get rather than a raw get.
    def test_get_json_file(self):
        data = self.instance.files.get("/tmp/json-get")
        self.assertEqual(b'{"some": "value"}', data)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from pylxd import exceptions, models
from pylxd.tests import testing


class TestOperation(testing.PyLXDTestCase):
    """Tests for pylxd.models.Operation."""

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": ""})
    @mock.patch("warnings.warn")
    def test_init_warnings_once(self, mock_warn):
        with mock.patch("pylxd.models.operation._seen_attribute_warnings", new=set()):
            models.Operation(unknown="some_value")
            mock_warn.assert_called_once_with(mock.ANY)
            models.Operation(unknown="some_value_as_well")
            mock_warn.assert_called_once_with(mock.ANY)
            models.Operation(unknown2="some_2nd_value")
            self.assertEqual(len(mock_warn.call_args_list), 2)

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": "none"})
    @mock.patch("warnings.warn")
    def test_init_warnings_none(self, mock_warn):
        with mock.patch("pylxd.models.operation._seen_attribute_warnings", new=set()):
            models.Operation(unknown="some_value")
            mock_warn.assert_not_called()

    @mock.patch.dict("os.environ", {"PYLXD_WARNINGS": "always"})
    @mock.patch("warnings.warn")
    def test_init_warnings_always(self, mock_warn):
        with mock.patch("pylxd.models.operation._seen_attribute_warnings", new=set()):
            models.Operation(unknown="some_value")
            mock_warn.assert_called_once_with(mock.ANY)
            models.Operation(unknown="some_value_as_well")
            self.assertEqual(len(mock_warn.call_args_list), 2)

    def test_get(self):
        """Return an operation."""
        name = "operation-abc"

        an_operation = models.Operation.get(self.client, name)

        self.assertEqual(name, an_operation.id)

    def test_get_full_path(self):
        """Return an operation even if the full path is specified."""
        name = "/1.0/operations/operation-abc"

        an_operation = models.Operation.get(self.client, name)

        self.assertEqual("operation-abc", an_operation.id)

    def test_get_full_path_and_project(self):
        """Return an operation even if the full path is specified."""
        name = "/1.0/operations/operation-abc?project=default"

        an_operation = models.Operation.get(self.client, name)

        self.assertEqual("operation-abc", an_operation.id)

    def test_wait_with_error(self):
        """If the operation errors, wait raises an exception."""

        def error(request, context):
            context.status_code = 200
            return {
                "type": "sync",
                "metadata": {
                    "status": "Failure",
                    "err": "Keep your foot off the blasted samoflange.",
                },
            }

        self.add_rule(
            {
                "json": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/operations/operation-abc/wait$",
            }
        )

        name = "/1.0/operations/operation-abc"

        an_operation = models.Operation.get(self.client, name)

        self.assertRaises(exceptions.LXDAPIException, an_operation.wait)

    def test_wait_with_error_async_without_metadata(self):
        """If the operation errors, wait raises an exception."""

        def error(request, context):
            context.status_code = 200
            return {"type": "async", "error": "Could not proceed", "metadata": None}

        self.add_rule(
            {
                "json": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/operations/operation-abc/wait$",
            }
        )

        name = "/1.0/operations/operation-abc"

        an_operation = models.Operation.get(self.client, name)

        with self.assertRaises(exceptions.LXDAPIException) as wait_cm:
            an_operation.wait()
        assert str(wait_cm.exception) == "Could not proceed"

    def test_unknown_attribute(self):
        self.add_rule(
            {
                "text": json.dumps(
                    {
                        "type": "sync",
                        "metadata": {
                            "id": "operation-unknown",
                            "metadata": {"return": 0},
                            "unknown": False,
                        },
                    }
                ),
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/operations/operation-unknown$",
            }
        )
        url = "/1.0/operations/operation-unknown"
        models.Operation.get(self.client, url)



# Copyright (c) 2016-2020 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import errno
import json
import logging
import os
import stat
import time
from typing import IO, NamedTuple
from urllib import parse

from ws4py.client import WebSocketBaseClient
from ws4py.manager import WebSocketManager
from ws4py.messaging import BinaryMessage

from pylxd import managers
from pylxd.exceptions import LXDAPIException
from pylxd.models import _model as model
from pylxd.models.operation import Operation


class InstanceState(model.AttributeDict):
    """A simple object for representing instance state."""


class _InstanceExecuteResult(NamedTuple):
    exit_code: int
    stdout: IO
    stderr: IO


class Instance(model.Model):
    """An LXD Instance.

    This class is not intended to be used directly, but rather to be used
    via `Client.instance.create`.
    """

    architecture = model.Attribute()
    config = model.Attribute()
    created_at = model.Attribute()
    devices = model.Attribute()
    ephemeral = model.Attribute()
    expanded_config = model.Attribute()
    expanded_devices = model.Attribute()
    name = model.Attribute(readonly=True)
    description = model.Attribute()
    profiles = model.Attribute()
    status = model.Attribute(readonly=True)
    last_used_at = model.Attribute(readonly=True)
    location = model.Attribute(readonly=True)
    type = model.Attribute(readonly=True)
    project = model.Attribute(readonly=True, optional=True)

    status_code = model.Attribute(readonly=True)
    stateful = model.Attribute(readonly=True)

    snapshots = model.Manager()
    files = model.Manager()

    _endpoint = "instances"

    def __setattr__(self, name, value):
        if name == "location" and not self.client.server_clustered:
            # LXD reports "none" as location when not in a cluster
            value = None
        super().__setattr__(name, value)

    @property
    def api(self):
        return self.client.api[self._endpoint][self.name]

    class FilesManager:
        """A pseudo-manager for namespacing file operations."""

        def __init__(self, instance):
            self._instance = instance
            self._endpoint = instance.client.api[instance._endpoint][
                instance.name
            ].files

        def put(self, filepath, data, mode=None, uid=None, gid=None):
            """Push a file to the instance.

            This pushes a single file to the instances file system named by
            the `filepath`.

            :param filepath: The path in the instance to to store the data in.
            :type filepath: str
            :param data: The data to store in the file.
            :type data: bytes or str
            :param mode: The unit mode to store the file with.  The default of
                None stores the file with the current mask of 0700, which is
                the lxd default. If mode is True, preserve the file permissions
                from filepath.
            :type mode: Union[oct, int, str, bool]
            :param uid: The uid to use inside the instance. Default of None
                results in 0 (root).
            :type uid: int
            :param gid: The gid to use inside the instance.  Default of None
                results in 0 (root).
            :type gid: int
            :raises: LXDAPIException if something goes wrong
            """
            if isinstance(mode, bool):
                mode = os.stat(filepath).st_mode if mode else None
            headers = self._resolve_headers(mode=mode, uid=uid, gid=gid)
            response = self._endpoint.post(
                params={"path": filepath}, data=data, headers=headers or None
            )
            if response.status_code == 200:
                return
            raise LXDAPIException(response)

        def mk_dir(self, path, mode=None, uid=None, gid=None):
            """Creates an empty directory on the container.
            This pushes an empty directory to the containers file system
            named by the `filepath`.

            :param path: The path in the container to to store the data in.
            :type path: str
            :param mode: The unit mode to store the file with.  The default of
                None stores the file with the current mask of 0700, which is
                the lxd default.
            :type mode: Union[oct, int, str]
            :param uid: The uid to use inside the container. Default of None
                results in 0 (root).
            :type uid: int
            :param gid: The gid to use inside the container.  Default of None
                results in 0 (root).
            :type gid: int
            :raises: LXDAPIException if something goes wrong
            """
            headers = self._resolve_headers(mode=mode, uid=uid, gid=gid)
            headers["X-LXD-type"] = "directory"
            response = self._endpoint.post(params={"path": path}, headers=headers)
            if response.status_code == 200:
                return
            raise LXDAPIException(response)

        @staticmethod
        def _resolve_headers(headers=None, mode=None, uid=None, gid=None):
            if headers is None:
                headers = {}
            if mode is not None:
                if isinstance(mode, int):
                    mode = format(mode, "o")
                if not isinstance(mode, str):
                    raise ValueError("'mode' parameter must be int or string")
                if not mode.startswith("0"):
                    mode = f"0{mode}"
                headers["X-LXD-mode"] = mode
            if uid is not None:
                headers["X-LXD-uid"] = str(uid)
            if gid is not None:
                headers["X-LXD-gid"] = str(gid)
            return headers

        def delete_available(self):
            """File deletion is an extension API and may not be available.
            https://documentation.ubuntu.com/lxd/en/latest/api-extensions/#file-delete
            """
            return self._instance.client.has_api_extension("file_delete")

        def delete(self, filepath):
            self._instance.client.assert_has_api_extension("file_delete")
            response = self._endpoint.delete(params={"path": filepath})
            if response.status_code != 200:
                raise LXDAPIException(response)

        def get(self, filepath):
            response = self._endpoint.get(params={"path": filepath}, is_api=False)
            return response.content

        def recursive_put(self, src, dst, mode=None, uid=None, gid=None):
            """Recursively push directory to the instance.

            Recursively pushes directory to the instances
            named by the `dst`

            :param src: The source path of directory to copy.
            :type src: str
            :param dst: The destination path in the instance
                    of directory to copy
            :type dst: str
            :param mode: The unit mode to store the file with.  The default of
                None stores the file with the current mask of 0700, which is
                the lxd default. If mode is True, preserve the file permissions
                from src.
            :type mode: Union[oct, int, str, bool]
            :param uid: The uid to use inside the instance. Default of None
                results in 0 (root).
            :type uid: int
            :param gid: The gid to use inside the instance.  Default of None
                results in 0 (root).
            :type gid: int
            :raises: NotADirectoryError if src is not a directory
            :raises: LXDAPIException if an error occurs
            """
            norm_src = os.path.normpath(src)
            if not os.path.isdir(norm_src):
                raise NotADirectoryError("'src' parameter must be a directory ")

            idx = len(norm_src)
            dst_items = set()

            for path, dirname, files in os.walk(norm_src):
                dst_path = os.path.normpath(
                    os.path.join(dst, path[idx:].lstrip(os.path.sep))
                )
                # create directory or symlink (depending on what's there)
                if path not in dst_items:
                    dst_items.add(path)
                    fmode = os.stat(path).st_mode
                    current_mode = mode
                    if isinstance(mode, bool):
                        current_mode = fmode if mode else None
                    headers = self._resolve_headers(mode=current_mode, uid=uid, gid=gid)
                    # determine what the file is: a directory or a symlink
                    if stat.S_ISLNK(fmode):
                        headers["X-LXD-type"] = "symlink"
                    else:
                        headers["X-LXD-type"] = "directory"
                    self._endpoint.post(params={"path": dst_path}, headers=headers)

                # copy files
                for f in files:
                    src_file = os.path.join(path, f)
                    with open(src_file, "rb") as fp:
                        filepath = os.path.join(dst_path, f)
                        current_mode = mode
                        if isinstance(mode, bool):
                            current_mode = os.stat(src_file).st_mode if mode else None
                        headers = self._resolve_headers(
                            mode=current_mode, uid=uid, gid=gid
                        )
                        response = self._endpoint.post(
                            params={"path": filepath},
                            data=fp.read(),
                            headers=headers or None,
                        )
                        if response.status_code != 200:
                            raise LXDAPIException(response)

        def recursive_get(self, remote_path, local_path):
            """Recursively pulls a directory from the container.
            Pulls the directory named `remote_path` from the container and
            creates a local folder named `local_path` with the
            content of `remote_path`.
            If `remote_path` is a file, it will be copied to `local_path`.

            :param remote_path: The directory path on the container.
            :type remote_path: str
            :param local_path: The path at which the directory will be stored.
            :type local_path: str
            :raises: LXDAPIException if an error occurs
            """
            response = self._endpoint.get(params={"path": remote_path}, is_api=False)

            if "X-LXD-type" in response.headers:
                if response.headers["X-LXD-type"] == "directory":
                    # TODO: We considered using the X-LXD-uid, X-LXD-gid,
                    #       and X-LXD-mode header information, but it was
                    #       beyond the scope of this change.
                    os.mkdir(local_path)
                    content = json.loads(response.content)
                    if "metadata" in content and content["metadata"]:
                        for file in content["metadata"]:
                            self.recursive_get(
                                os.path.join(remote_path, file),
                                os.path.join(local_path, file),
                            )
                elif response.headers["X-LXD-type"] == "file":
                    with open(local_path, "wb") as f:
                        # TODO: Same thoughts on file permissions as above.
                        f.write(response.content)

    @classmethod
    def exists(cls, client, name):
        """Determine whether a instance exists."""
        try:
            getattr(client, cls._endpoint).get(name)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, name):
        """Get a instance by name."""
        response = client.api[cls._endpoint][name].get()

        return cls(client, **response.json()["metadata"])

    @classmethod
    def all(cls, client, recursion=0):
        """Get all instances.

        This method returns an Instance array. If recursion is unset,
        only the name of each instance will be set and `Instance.sync`
        can be used to return more information. If recursion is between
        1-2 this method will pre-fetch additional instance attributes for
        all instances in the array.
        """
        params = {}
        if recursion != 0:
            params = {"recursion": recursion}
        response = client.api[cls._endpoint].get(params=params)

        instances = []
        for instance in response.json()["metadata"]:
            if isinstance(instance, dict):
                # User specified recursion so returning all data for each instance at once
                instance_class = cls(client, name=instance["name"])
                for key, data in instance.items():
                    try:
                        setattr(instance_class, key, data)
                    except AttributeError:
                        pass
                instances.append(instance_class)
            else:
                name = instance.split("/")[-1]
                instances.append(cls(client, name=name))
        return instances

    @classmethod
    def create(cls, client, config, wait=False, target=None):
        """Create a new instance config.

        :param client: client instance
        :type client: Client
        :param config: The configuration for the new instance.
        :type config: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :param target: If in cluster mode, the target member.
        :type target: str
        :raises LXDAPIException: if something goes wrong.
        :returns: an instance if successful
        :rtype: :class:`Instance`
        """
        response = client.api[cls._endpoint].post(json=config, target=target)

        if wait:
            client.operations.wait_for_operation(response.json()["operation"])
        return cls(client, name=config["name"])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.snapshots = managers.SnapshotManager(self.client, self)
        self.files = self.FilesManager(self)

    def rename(self, name, wait=False):
        """Rename an instance."""
        response = self.api.post(json={"name": name})

        if wait:
            self.client.operations.wait_for_operation(response.json()["operation"])
        self.name = name

    def _set_state(self, state, timeout=30, force=True, wait=False):
        response = self.api.state.put(
            json={"action": state, "timeout": timeout, "force": force}
        )
        if wait:
            self.client.operations.wait_for_operation(response.json()["operation"])
            if "status" in self.__dirty__:
                self.__dirty__.remove("status")
            if self.ephemeral and state == "stop":
                self.client = None
            else:
                self.sync()

    def state(self):
        response = self.api.state.get()
        state = InstanceState(response.json()["metadata"])
        return state

    def start(self, timeout=30, force=True, wait=False):
        """Start the instance."""
        return self._set_state("start", timeout=timeout, force=force, wait=wait)

    def stop(self, timeout=30, force=True, wait=False):
        """Stop the instance."""
        return self._set_state("stop", timeout=timeout, force=force, wait=wait)

    def restart(self, timeout=30, force=True, wait=False):
        """Restart the instance."""
        return self._set_state("restart", timeout=timeout, force=force, wait=wait)

    def freeze(self, timeout=30, force=True, wait=False):
        """Freeze the instance."""
        return self._set_state("freeze", timeout=timeout, force=force, wait=wait)

    def unfreeze(self, timeout=30, force=True, wait=False):
        """Unfreeze the instance."""
        return self._set_state("unfreeze", timeout=timeout, force=force, wait=wait)

    def execute(
        self,
        commands,
        environment=None,
        encoding=None,
        decode=True,
        stdin_payload=None,
        stdin_encoding="utf-8",
        stdout_handler=None,
        stderr_handler=None,
        user=None,
        group=None,
        cwd=None,
    ):
        """Execute a command on the instance. stdout and stderr are buffered if
        no handler is given.

        :param commands: The command and arguments as a list of strings
        :type commands: [str]
        :param environment: The environment variables to pass with the command
        :type environment: {str: str}
        :param encoding: The encoding to use for stdout/stderr if the param
            decode is True.  If encoding is None, then no override is
            performed and whatever the existing encoding from LXD is used.
        :type encoding: str
        :param decode: Whether to decode the stdout/stderr or just return the
            raw buffers.
        :type decode: bool
        :param stdin_payload: Payload to pass via stdin
        :type stdin_payload: Can be a file, string, bytearray, generator or
            ws4py Message object
        :param stdin_encoding: Encoding to pass text to stdin (default utf-8)
        :param stdout_handler: Callable than receive as first parameter each
            message received via stdout
        :type stdout_handler: Callable[[str], None]
        :param stderr_handler: Callable than receive as first parameter each
            message received via stderr
        :type stderr_handler: Callable[[str], None]
        :param user: User to run the command as
        :type user: int
        :param group: Group to run the command as
        :type group: int
        :param cwd: Current working directory
        :type cwd: str
        :returns: A tuple of `(exit_code, stdout, stderr)`
        :rtype: _InstanceExecuteResult() namedtuple
        """
        if isinstance(commands, str):
            raise TypeError("First argument must be a list.")
        if environment is None:
            environment = {}

        response = self.api["exec"].post(
            json={
                "command": commands,
                "environment": environment,
                "wait-for-websocket": True,
                "interactive": False,
                "user": user,
                "group": group,
                "cwd": cwd,
            }
        )

        response_json = response.json()
        fds = response_json["metadata"]["metadata"]["fds"]
        operation_id = Operation.extract_operation_id(response_json["operation"])
        parsed = parse.urlparse(
            self.client.api.operations[operation_id].websocket._api_endpoint
        )

        with managers.web_socket_manager(WebSocketManager()) as manager:
            stdin = _StdinWebsocket(
                self.client.websocket_url,
                payload=stdin_payload,
                encoding=stdin_encoding,
                ssl_options=self.client.ssl_options,
            )
            stdin.resource = f"{parsed.path}?secret={fds['0']}"
            stdin.connect()
            stdout = _CommandWebsocketClient(
                manager,
                self.client.websocket_url,
                encoding=encoding,
                decode=decode,
                handler=stdout_handler,
                ssl_options=self.client.ssl_options,
            )
            stdout.resource = f"{parsed.path}?secret={fds['1']}"
            stdout.connect()
            stderr = _CommandWebsocketClient(
                manager,
                self.client.websocket_url,
                encoding=encoding,
                decode=decode,
                handler=stderr_handler,
                ssl_options=self.client.ssl_options,
            )
            stderr.resource = f"{parsed.path}?secret={fds['2']}"
            stderr.connect()

            manager.start()

            # watch for the end of the command:
            while True:
                operation = self.client.operations.get(operation_id)
                if "return" in operation.metadata:
                    break
                time.sleep(0.5)  # pragma: no cover

            try:
                stdin.close()
            except BrokenPipeError:
                pass

            stdout.finish_soon()
            stderr.finish_soon()
            try:
                manager.close_all()
            except BrokenPipeError:
                pass

            while not stdout.finished or not stderr.finished:
                time.sleep(0.1)  # progma: no cover

            manager.stop()
            manager.join()

            return _InstanceExecuteResult(
                operation.metadata["return"], stdout.data, stderr.data
            )

    def raw_interactive_execute(
        self, commands, environment=None, user=None, group=None, cwd=None
    ):
        """Execute a command on the instance interactively and returns
        urls to websockets. The urls contain a secret uuid, and can be accesses
        without further authentication. The caller has to open and manage
        the websockets themselves.

        :param commands: The command and arguments as a list of strings
           (most likely a shell)
        :type commands: [str]
        :param environment: The environment variables to pass with the command
        :type environment: {str: str}
        :param user: User to run the command as
        :type user: int
        :param group: Group to run the command as
        :type group: int
        :param cwd: Current working directory
        :type cwd: str
        :returns: Two urls to an interactive websocket and a control socket
        :rtype: {'ws':str,'control':str}
        """
        if isinstance(commands, str):
            raise TypeError("First argument must be a list.")

        if environment is None:
            environment = {}

        response = self.api["exec"].post(
            json={
                "command": commands,
                "environment": environment,
                "wait-for-websocket": True,
                "interactive": True,
                "user": user,
                "group": group,
                "cwd": cwd,
            }
        )

        response_json = response.json()
        fds = response_json["metadata"]["metadata"]["fds"]
        operation_id = response_json["operation"].split("/")[-1].split("?")[0]
        parsed = parse.urlparse(
            self.client.api.operations[operation_id].websocket._api_endpoint
        )

        return {
            "ws": f"{parsed.path}?secret={fds['0']}",
            "control": f"{parsed.path}?secret={fds['control']}",
        }

    def migrate(self, new_client, live=False, wait=False):
        """Migrate a instance.

        Destination host information is contained in the client
        connection passed in.

        If the `live` param is True, then a live migration is attempted,
        otherwise a non live one is running.

        If the instance is running for live migration, it either must be shut
        down first or criu must be installed on the source and destination
        machines and the `live` param should be True.

        :param new_client: the pylxd client connection to migrate the instance
            to.
        :type new_client: :class:`pylxd.client.Client`
        :param live: whether to perform a live migration
        :type live: bool
        :param wait: if True, wait for the migration to complete
        :type wait: bool
        :raises: LXDAPIException if any of the API calls fail.
        :raises: ValueError if source of target are local connections
        :returns: the response from LXD of the new instance (the target of the
            migration and not the operation if waited on.)
        :rtype: :class:`requests.Response`
        """
        if self.api.scheme in ("http+unix",):
            raise ValueError("Cannot migrate from a local client connection")

        if self.status_code == 103:
            try:
                res = getattr(new_client, self._endpoint).create(
                    self.generate_migration_data(live), wait=wait
                )
            except LXDAPIException as e:
                if e.response.status_code == 103:
                    self.delete()
                    return getattr(new_client, self._endpoint).get(self.name)
                else:
                    raise e
        else:
            res = getattr(new_client, self._endpoint).create(
                self.generate_migration_data(live), wait=wait
            )
        self.delete()
        return res

    def generate_migration_data(self, live=False):
        """Generate the migration data.

        This method can be used to handle migrations where the client
        connection uses the local unix socket. For more information on
        migration, see `Instance.migrate`.

        :param live: Whether to include "live": "true" in the migration
        :type live: bool
        :raises: LXDAPIException if the request to migrate fails
        :returns: dictionary of migration data suitable to send to an new
            client to complete a migration.
        :rtype: Dict[str, ANY]
        """
        self.sync()  # Make sure the object isn't stale
        _json = {"migration": True}
        if live:
            _json["live"] = True
        response = self.api.post(json=_json)
        response_json = response.json()
        operation = self.client.operations.get(response_json["operation"])
        operation_url = self.client.api.operations[operation.id]._api_endpoint
        secrets = response_json["metadata"]["metadata"]
        cert = self.client.host_info["environment"]["certificate"]

        return {
            "name": self.name,
            "architecture": self.architecture,
            "config": self.config,
            "devices": self.devices,
            "epehemeral": self.ephemeral,
            "default": self.profiles,
            "source": {
                "type": "migration",
                "operation": operation_url,
                "mode": "pull",
                "certificate": cert,
                "secrets": secrets,
            },
        }

    def publish(self, public=False, wait=False):
        """Publish a instance as an image.

        The instance must be stopped in order publish it as an image. This
        method does not enforce that constraint, so a LXDAPIException may be
        raised if this method is called on a running instance.

        If wait=True, an Image is returned.
        """
        data = {
            "public": public,
            "source": {
                "type": self.type,
                "name": self.name,
            },
        }

        response = self.client.api.images.post(json=data)
        if wait:
            operation = self.client.operations.wait_for_operation(
                response.json()["operation"]
            )

            return self.client.images.get(operation.metadata["fingerprint"])

    def restore_snapshot(self, snapshot_name, wait=False, stateful=False):
        """Restore a snapshot using its name.

        Attempts to restore a instance using a snapshot previously made.  The
        instance should be stopped, but the method does not enforce this
        constraint, so an LXDAPIException may be raised if this method fails.

        :param snapshot_name: the name of the snapshot to restore from
        :type snapshot_name: str
        :param wait: wait until the operation is completed.
        :type wait: boolean
        :param stateful: restore the instance's running state from the snapshot (if
            available), set this to ``True`` if the snapshot is stateful
        :type stateful: boolean
        :raises: LXDAPIException if the the operation fails.
        :returns: the original response from the restore operation (not the
            operation result)
        :rtype: :class:`requests.Response`
        """
        response = self.api.put(json={"restore": snapshot_name, "stateful": stateful})
        if wait:
            self.client.operations.wait_for_operation(response.json()["operation"])
        return response


class _CommandWebsocketClient(WebSocketBaseClient):  # pragma: no cover
    """Handle a websocket for instance.execute(...) and manage decoding of the
    returned values depending on 'decode' and 'encoding' parameters.
    """

    def __init__(self, manager, *args, **kwargs):
        self.manager = manager
        self.decode = kwargs.pop("decode", True)
        self.encoding = kwargs.pop("encoding", None)
        self.handler = kwargs.pop("handler", None)
        self.message_encoding = None
        self.finish_off = False
        self.finished = False
        self.last_message_empty = False
        self.buffer = []
        super().__init__(*args, **kwargs)

    def handshake_ok(self):
        self.manager.add(self)
        self.buffer = []

    def received_message(self, message):
        if message.data is None or len(message.data) == 0:
            self.last_message_empty = True
            if self.finish_off:
                self.finished = True
            return
        else:
            self.last_message_empty = False
        if message.encoding and self.message_encoding is None:
            self.message_encoding = message.encoding
        if self.handler:
            self.handler(self._maybe_decode(message.data))
        else:
            self.buffer.append(message.data)
        if self.finish_off and isinstance(message, BinaryMessage):
            self.finished = True

    def closed(self, code, reason=None):
        self.finished = True

    def finish_soon(self):
        self.finish_off = True
        if self.last_message_empty:
            self.finished = True

    def _maybe_decode(self, buffer):
        if self.decode and buffer is not None:
            if self.encoding:
                return buffer.decode(self.encoding)
            if self.message_encoding:
                return buffer.decode(self.message_encoding)
            # This is the backwards compatible "always decode to utf-8"
            return buffer.decode("utf-8")
        return buffer

    @property
    def data(self):
        buffer = b"".join(self.buffer)
        return self._maybe_decode(buffer)

    def unhandled_error(self, error):
        """
        Handles the unfriendly socket closures on the server side
        without showing a confusing error message
        """
        if hasattr(error, "errno") and error.errno == errno.ECONNRESET:
            pass
        else:
            logger = logging.getLogger("ws4py")
            logger.exception("Failed to receive data")


class _StdinWebsocket(WebSocketBaseClient):  # pragma: no cover
    """A websocket client for handling stdin.

    Allow comunicate with instance commands via stdin
    """

    def __init__(self, url, payload=None, **kwargs):
        self.encoding = kwargs.pop("encoding", None)
        self.payload = payload
        super().__init__(url, **kwargs)

    def _smart_encode(self, msg):
        if isinstance(msg, str) and self.encoding:
            return msg.encode(self.encoding)
        return msg

    def handshake_ok(self):
        if self.payload:
            if hasattr(self.payload, "read"):
                self.send(
                    (self._smart_encode(line) for line in self.payload), binary=True
                )
            else:
                self.send(self._smart_encode(self.payload), binary=True)
        self.send(b"", binary=False)


class Snapshot(model.Model):
    """A instance snapshot."""

    name = model.Attribute()
    created_at = model.Attribute()
    stateful = model.Attribute()

    instance = model.Parent()

    @property
    def api(self):
        return self.client.api[self.instance._endpoint][self.instance.name].snapshots[
            self.name
        ]

    @classmethod
    def get(cls, client, instance, name):
        response = client.api[instance._endpoint][instance.name].snapshots[name].get()

        snapshot = cls(client, instance=instance, **response.json()["metadata"])
        # Snapshot names are namespaced in LXD, as
        # instance-name/snapshot-name. We hide that implementation
        # detail.
        snapshot.name = snapshot.name.split("/")[-1]
        return snapshot

    @classmethod
    def all(cls, client, instance):
        response = client.api[instance._endpoint][instance.name].snapshots.get()

        return [
            cls(client, name=snapshot.split("/")[-1], instance=instance)
            for snapshot in response.json()["metadata"]
        ]

    @classmethod
    def create(cls, client, instance, name, stateful=False, wait=False):
        response = client.api[instance._endpoint][instance.name].snapshots.post(
            json={"name": name, "stateful": stateful}
        )

        snapshot = cls(client, instance=instance, name=name)
        if wait:
            client.operations.wait_for_operation(response.json()["operation"])
        return snapshot

    def rename(self, new_name, wait=False):
        """Rename a snapshot."""
        response = self.api.post(json={"name": new_name})
        if wait:
            self.client.operations.wait_for_operation(response.json()["operation"])
        self.name = new_name

    def publish(self, public=False, wait=False):
        """Publish a snapshot as an image.

        If wait=True, an Image is returned.
        """
        data = {
            "public": public,
            "source": {
                "type": "snapshot",
                "name": f"{self.instance.name}/{self.name}",
            },
        }

        response = self.client.api.images.post(json=data)
        if wait:
            operation = self.client.operations.wait_for_operation(
                response.json()["operation"]
            )
            return self.client.images.get(operation.metadata["fingerprint"])

    def restore(self, wait=False):
        """Restore this snapshot.

        Attempts to restore a instance using this snapshot.  The instance
        should be stopped, but the method does not enforce this constraint, so
        an LXDAPIException may be raised if this method fails.

        :param wait: wait until the operation is completed.
        :type wait: boolean
        :raises: LXDAPIException if the the operation fails.
        :returns: the original response from the restore operation (not the
            operation result)
        :rtype: :class:`requests.Response`
        """
        return self.instance.restore_snapshot(
            self.name, wait, stateful=getattr(self, "stateful", False)
        )



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from pylxd import exceptions, models
from pylxd.exceptions import LXDAPIExtensionNotAvailable
from pylxd.tests import testing


class TestNetwork(testing.PyLXDTestCase):
    """Tests for pylxd.models.Network."""

    def test_get(self):
        """A network is fetched."""
        name = "eth0"
        an_network = models.Network.get(self.client, name)

        self.assertEqual(name, an_network.name)

    def test_get_not_found(self):
        """LXDAPIException is raised on unknown network."""

        def not_found(_, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException, models.Network.get, self.client, "eth0"
        )

    def test_get_error(self):
        """LXDAPIException is raised on error."""

        def error(_, context):
            context.status_code = 500
            return json.dumps(
                {
                    "type": "error",
                    "error": "Not found",
                    "error_code": 500,
                }
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException, models.Network.get, self.client, "eth0"
        )

    def test_exists(self):
        """True is returned if network exists."""
        name = "eth0"

        self.assertTrue(models.Network.exists(self.client, name))

    def test_not_exists(self):
        """False is returned when network does not exist."""

        def not_found(_, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0$",
            }
        )

        name = "eth0"

        self.assertFalse(models.Network.exists(self.client, name))

    def test_all(self):
        networks = models.Network.all(self.client)

        self.assertEqual(2, len(networks))

    def test_create_with_parameters(self):
        with mock.patch.object(self.client, "assert_has_api_extension"):
            network = models.Network.create(
                self.client,
                name="eth1",
                config={},
                type="bridge",
                description="Network description",
            )

        self.assertIsInstance(network, models.Network)
        self.assertEqual("eth1", network.name)
        self.assertEqual("Network description", network.description)
        self.assertEqual("bridge", network.type)
        self.assertTrue(network.managed)

    def test_create_default(self):
        with mock.patch.object(self.client, "assert_has_api_extension"):
            network = models.Network.create(self.client, "eth1")

        self.assertIsInstance(network, models.Network)
        self.assertEqual("eth1", network.name)
        self.assertEqual("bridge", network.type)
        self.assertTrue(network.managed)

    def test_create_api_not_available(self):
        # Note, by default with the tests, no 'network' extension is available.
        with self.assertRaises(LXDAPIExtensionNotAvailable):
            models.Network.create(
                self.client,
                name="eth1",
                config={},
                type="bridge",
                description="Network description",
            )

    def test_rename(self):
        with mock.patch.object(self.client, "assert_has_api_extension"):
            network = models.Network.get(self.client, "eth0")
            renamed_network = network.rename("eth2")

        self.assertEqual("eth2", renamed_network.name)

    def test_update(self):
        """A network is updated."""
        with mock.patch.object(self.client, "assert_has_api_extension"):
            network = models.Network.get(self.client, "eth0")
            network.config = {}
            network.save()
        self.assertEqual({}, network.config)

    def test_fetch(self):
        """A partial network is synced."""
        network = self.client.networks.all()[1]

        network.sync()

        self.assertEqual("Network description", network.description)

    def test_fetch_not_found(self):
        """LXDAPIException is raised on bogus network fetch."""

        def not_found(_, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0$",
            }
        )
        network = models.Network(self.client, name="eth0")

        self.assertRaises(exceptions.LXDAPIException, network.sync)

    def test_fetch_error(self):
        """LXDAPIException is raised on fetch error."""

        def error(_, context):
            context.status_code = 500
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 500}
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0$",
            }
        )
        network = models.Network(self.client, name="eth0")

        self.assertRaises(exceptions.LXDAPIException, network.sync)

    def test_delete(self):
        """A network is deleted."""
        network = models.Network(self.client, name="eth0")

        network.delete()

    def test_state(self):
        state = {
            "addresses": [
                {
                    "family": "inet",
                    "address": "10.87.252.1",
                    "netmask": "24",
                    "scope": "global",
                },
                {
                    "family": "inet6",
                    "address": "fd42:6e0e:6542:a212::1",
                    "netmask": "64",
                    "scope": "global",
                },
            ],
            "counters": {
                "bytes_received": 0,
                "bytes_sent": 17724,
                "packets_received": 0,
                "packets_sent": 95,
            },
            "hwaddr": "36:19:09:9b:f9:aa",
            "mtu": 1500,
            "state": "up",
            "type": "broadcast",
        }
        self.add_rule(
            {
                "json": {
                    "type": "sync",
                    "metadata": state,
                },
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0/state$",
            }
        )
        network = models.Network.get(self.client, "eth0")
        assert dict(network.state()) == state

    def test_str(self):
        """Network is printed in JSON format."""
        network = models.Network.get(self.client, "eth0")
        self.assertEqual(
            json.loads(str(network)),
            {
                "name": "eth0",
                "description": "Network description",
                "type": "bridge",
                "config": {
                    "ipv4.address": "10.80.100.1/24",
                    "ipv4.nat": "true",
                    "ipv6.address": "none",
                    "ipv6.nat": "false",
                },
                "managed": True,
                "used_by": [],
            },
        )

    def test_repr(self):
        network = models.Network.get(self.client, "eth0")
        self.assertEqual(
            repr(network),
            'Network(config={"ipv4.address": "10.80.100.1/24", "ipv4.nat": '
            '"true", "ipv6.address": "none", "ipv6.nat": "false"}, '
            'description="Network description", name="eth0", type="bridge")',
        )


class TestNetworkForward(testing.PyLXDTestCase):
    """Tests for pylxd.models.NetworkForward."""

    def test_get(self):
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)
        forward = an_network.forwards.get("192.0.2.1")

        self.assertEqual("Forward description", forward.description)

    def test_get_not_found(self):
        """LXDAPIException is raised on unknown network."""
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)

        def not_found(_, context):
            context.status_code = 404
            return json.dumps(
                {"type": "error", "error": "Not found", "error_code": 404}
            )

        self.add_rule(
            {
                "text": not_found,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0/forwards/192.0.2.1$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException,
            models.NetworkForward.get,
            self.client,
            an_network,
            "192.0.2.1",
        )

    def test_get_error(self):
        """LXDAPIException is raised on error."""
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)

        def error(_, context):
            context.status_code = 500
            return json.dumps(
                {
                    "type": "error",
                    "error": "Not found",
                    "error_code": 500,
                }
            )

        self.add_rule(
            {
                "text": error,
                "method": "GET",
                "url": r"^http://pylxd.test/1.0/networks/eth0/forwards/192.0.2.1$",
            }
        )

        self.assertRaises(
            exceptions.LXDAPIException,
            models.NetworkForward.get,
            self.client,
            an_network,
            "192.0.2.1",
        )

    def test_create(self):
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)
        config = {
            "config": {},
            "description": "Forward description",
            "listen_address": "192.0.2.1",
            "ports": [
                {
                    "description": "Port description",
                    "listen_port": "80",
                    "target_address": "192.0.2.2",
                    "target_port": "80",
                }
            ],
        }

        with mock.patch.object(self.client, "assert_has_api_extension"):
            forward = models.NetworkForward.create(
                self.client,
                network=an_network,
                config=config,
            )

        self.assertIsInstance(forward, models.NetworkForward)
        self.assertEqual("Forward description", forward.description)
        self.assertEqual("192.0.2.1", forward.listen_address)
        self.assertEqual(len(forward.ports), 1)
        self.assertEqual("80", forward.ports[0]["listen_port"])
        self.assertEqual("192.0.2.2", forward.ports[0]["target_address"])
        self.assertEqual("80", forward.ports[0]["target_port"])
        self.assertEqual("Port description", forward.ports[0]["description"])

    def test_update(self):
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)
        config = {
            "config": {},
            "description": "Forward description",
            "listen_address": "192.0.2.1",
            "ports": [
                {
                    "description": "Port description",
                    "listen_port": "80",
                    "target_address": "192.0.2.2",
                    "target_port": "80",
                }
            ],
        }

        with mock.patch.object(self.client, "assert_has_api_extension"):
            forward = models.NetworkForward.create(
                self.client,
                network=an_network,
                config=config,
            )
            forward.description = "Updated"
            forward.save()

        self.assertIsInstance(forward, models.NetworkForward)
        self.assertEqual("Updated", forward.description)
        self.assertEqual("192.0.2.1", forward.listen_address)
        self.assertEqual(len(forward.ports), 1)
        self.assertEqual("80", forward.ports[0]["listen_port"])
        self.assertEqual("192.0.2.2", forward.ports[0]["target_address"])
        self.assertEqual("80", forward.ports[0]["target_port"])
        self.assertEqual("Port description", forward.ports[0]["description"])

    def test_str(self):
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)
        forward = an_network.forwards.get("192.0.2.1")
        self.assertEqual(
            json.loads(str(forward)),
            {
                "config": {},
                "description": "Forward description",
                "location": "eth0",
                "listen_address": "192.0.2.1",
                "ports": [
                    {
                        "description": "Port description",
                        "listen_port": "80",
                        "target_address": "192.0.2.2",
                        "target_port": "80",
                    }
                ],
            },
        )

    def test_repr(self):
        network_name = "eth0"
        an_network = models.Network.get(self.client, network_name)
        forward = an_network.forwards.get("192.0.2.1")
        self.assertEqual(
            repr(forward),
            'NetworkForward(config={}, description="Forward description", listen_address="192.0.2.1",'
            ' location="eth0", ports=[{"description": "Port description", "listen_port": "80",'
            ' "target_address": "192.0.2.2", "target_port": "80"}])',
        )



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import pylxd.exceptions as exceptions
from integration.testing import IntegrationTestCase


class StorageTestCase(IntegrationTestCase):
    def setUp(self):
        super().setUp()

        if not self.client.has_api_extension("storage"):
            self.skipTest("Required 'storage' LXD API extension not available!")


class TestStoragePools(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pools.StoragePools"""

    # note create and delete are tested in every method

    def test_get(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        self.assertEqual(name, storage_pool.name)

    def test_all(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pools = self.client.storage_pools.all()
        self.assertIn(name, [p.name for p in storage_pools])

    def test_save(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        storage_pool.description = "My storage pool"
        storage_pool.save()

        p = self.client.storage_pools.get(name)
        self.assertEqual(p.description, "My storage pool")

    def test_put(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        new_desc = "new description"
        self.assertNotEqual(storage_pool.description, new_desc)
        put_object = {
            "description": new_desc,
            "config": storage_pool.config,
        }
        storage_pool.put(put_object)
        self.assertEqual(storage_pool.description, new_desc)
        p = self.client.storage_pools.get(name)
        self.assertEqual(p.description, new_desc)

    def test_patch(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        new_desc = "new description"
        self.assertNotEqual(storage_pool.description, new_desc)
        patch_object = {
            "description": new_desc,
        }
        storage_pool.patch(patch_object)
        self.assertEqual(storage_pool.description, new_desc)
        p = self.client.storage_pools.get(name)
        self.assertEqual(p.description, new_desc)


class TestStorageResources(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pools.StorageResources"""

    def test_get(self):
        name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, name)

        storage_pool = self.client.storage_pools.get(name)
        # just assert that it can be fetched and that a key exists
        resources = storage_pool.resources.get()
        self.assertIsInstance(resources.space, dict)
        self.assertIsInstance(resources.inodes, dict)


class TestStorageVolume(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pools.StorageVolume"""

    def test_create_and_get_and_delete(self):
        pool_name = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, pool_name)
        storage_pool = self.client.storage_pools.get(pool_name)

        volume = self.create_storage_volume(pool_name, "vol1")
        vol_copy = storage_pool.volumes.get("custom", "vol1")
        self.assertEqual(vol_copy.name, volume.name)
        volume.delete(wait=True)

    @unittest.skip("Can't test PUT on volumes as it doesn't make sense yet")
    def test_put(self):
        pass

    @unittest.skip("Can't test .save() on volumes yet - doesn't make sense")
    def test_save(self):
        pass

    @unittest.skip("Can't test PATCH on volumes yet - doesn't make sense")
    def test_patch(self):
        # as we're not using ZFS (and can't in these integration tests) we
        # can't really patch anything on a dir volume.
        pass


class TestStorageVolumeSnapshot(StorageTestCase):
    """Tests for :py:class:`pylxd.models.storage_pool.StorageVolumeSnapshot"""

    def setUp(self):
        super().setUp()

        if not self.client.has_api_extension("storage_api_volume_snapshots"):
            self.skipTest(
                "Required 'storage_api_volume_snapshots' LXD API extension not available!"
            )

    def test_create_get_restore_delete_volume_snapshot(self):
        # Create pool and volume
        pool = self.create_storage_pool()
        self.addCleanup(self.delete_storage_pool, pool)

        volume = self.create_storage_volume(pool, "vol1")
        self.addCleanup(self.delete_storage_volume, pool, "vol1")

        # Create a few snapshots
        first_snapshot = volume.snapshots.create(wait=True)
        self.assertEqual(first_snapshot.name, "snap0")

        second_snapshot = volume.snapshots.create(wait=True)
        self.assertEqual(second_snapshot.name, "snap1")

        # Try restoring the volume from one of the snapshots
        first_snapshot.restore(wait=True)

        # Create new snapshot with defined name and expiration date
        custom_snapshot_name = "custom-snapshot"
        custom_snapshot_expiry_date = "2183-06-16T00:00:00Z"

        custom_snapshot = volume.snapshots.create(
            name=custom_snapshot_name, expires_at=custom_snapshot_expiry_date, wait=True
        )
        self.assertEqual(custom_snapshot.name, custom_snapshot_name)
        self.assertEqual(custom_snapshot.expires_at, custom_snapshot_expiry_date)

        # Get all snapshots from the volume
        all_snapshots = volume.snapshots.all()
        self.assertEqual(len(all_snapshots), 3)

        for snapshot_name in ["snap0", "snap1", custom_snapshot_name]:
            self.assertIn(snapshot_name, all_snapshots)

        # Delete a snapshot
        second_snapshot.delete(wait=True)
        self.assertFalse(volume.snapshots.exists(second_snapshot.name))

        self.assertRaises(exceptions.NotFound, volume.snapshots.get, "snap1")

        all_snapshots = volume.snapshots.all()
        self.assertEqual(len(all_snapshots), 2)

        for snapshot_name in ["snap0", custom_snapshot_name]:
            self.assertIn(snapshot_name, all_snapshots)

        self.assertFalse("snap1" in all_snapshots)

        # Test getting all snapshots with recursion
        all_snapshots = volume.snapshots.all(use_recursion=True)
        self.assertEqual(first_snapshot, all_snapshots[0])
        self.assertEqual(custom_snapshot, all_snapshots[1])

        # Change snapshot values
        first_snapshot.rename("first", wait=True)
        self.assertFalse(volume.snapshots.exists("snap0"))
        self.assertRaises(exceptions.NotFound, volume.snapshots.get, "snap0")

        new_description = "first snapshot"
        first_snapshot.description = new_description
        first_snapshot.save(wait=True)
        self.assertEqual(volume.snapshots.get("first").description, new_description)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from pylxd.models import _model as model


class Profile(model.Model):
    """A LXD profile."""

    config = model.Attribute()
    description = model.Attribute()
    devices = model.Attribute()
    name = model.Attribute(readonly=True)
    used_by = model.Attribute(readonly=True)

    @classmethod
    def exists(cls, client, name):
        """Determine whether a profile exists."""
        try:
            client.profiles.get(name)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, name):
        """Get a profile."""
        response = client.api.profiles[name].get()
        return cls(client, **response.json()["metadata"])

    @classmethod
    def all(cls, client):
        """Get all profiles."""
        response = client.api.profiles.get()

        profiles = []
        for url in response.json()["metadata"]:
            name = url.split("/")[-1]
            name = name.split("?")[0]
            profiles.append(cls(client, name=name))
        return profiles

    @classmethod
    def create(
        cls, client, name, config=None, devices=None, description=None, wait=False
    ):
        """Create a profile.

        :param client: The pylxd client object
        :type client: :class:`pylxd.client.Client`
        :param name: The name of the profile to create
        :type name: str
        :param config: Configuration dictionary for the profile
        :type config: dict
        :param devices: Devices dictionary for the profile
        :type devices: dict
        :param description: Description of the profile
        :type description: str
        :param wait: Whether to wait for async operations to complete
        :type wait: bool
        :returns: The created profile
        :rtype: :class:`pylxd.models.profile.Profile`
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the profile
            couldn't be created.
        """
        profile = {"name": name}
        if config is not None:
            profile["config"] = config
        if devices is not None:
            profile["devices"] = devices
        if description is not None:
            profile["description"] = description

        response = client.api.profiles.post(json=profile)
        # Handle async response if needed
        cls._handle_async_response_for_client(client, response, wait)
        return cls.get(client, name)

    @property
    def api(self):
        return self.client.api.profiles[self.name]

    def rename(self, new_name, wait=False):
        """Rename the profile.

        :param new_name: The new name for the profile
        :type new_name: str
        :param wait: Whether to wait for async operations to complete
        :type wait: bool
        :returns: The renamed profile
        :rtype: :class:`pylxd.models.profile.Profile`
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the profile
            couldn't be renamed.
        """
        response = self.api.post(json={"name": new_name})

        # Handle async response if needed
        self._handle_async_response(response, wait)
        return Profile.get(self.client, new_name)

    def save(self, wait=False):
        """Save the profile using PUT back to the LXD server.

        Implements PUT /1.0/profiles/<self.name>

        The fields affected are: `config`, `description`, and `devices`.
        Note that they are replaced in their *entirety*. If finer grained
        control is required, please use the
        :meth:`~pylxd.models.profile.Profile.patch` method directly.

        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the profile
            couldn't be saved.
        """
        # Note: This overrides the base save method to add documentation
        # but uses the parent's implementation which now handles async responses
        super().save(wait=wait)

    def put(self, put_object, wait=False):
        """Put the profile.

        Implements PUT /1.0/profiles/<self.name>

        Putting to a profile may fail if the new configuration is
        incompatible. See the LXD documentation for further details.

        Note that the object is refreshed with a `sync` if the PUT is
        successful. If this is *not* desired, then the raw API on the client
        should be used.

        :param put_object: A dictionary containing profile configuration.
        :type put_object: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the profile
            couldn't be modified.
        """
        # Note: This overrides the base put method to add documentation
        # but uses the parent's implementation which now handles async responses
        super().put(put_object, wait)

    def patch(self, patch_object, wait=False):
        """Patch the profile.

        Implements PATCH /1.0/profiles/<self.name>

        Patching the object allows for more fine grained changes to the config.
        The object is refreshed if the PATCH is successful. If this is *not*
        required, then use the client api directly.

        :param patch_object: A dictionary containing profile configuration updates.
        :type patch_object: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the profile
            couldn't be modified.
        """
        # Note: This overrides the base patch method to add documentation
        # but uses the parent's implementation which now handles async responses
        super().patch(patch_object, wait)

    def delete(self, wait=False):
        """Delete the profile.

        Implements DELETE /1.0/profiles/<self.name>

        Deleting a profile may fail if it is still in use.

        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the profile
            couldn't be deleted.
        """
        # Note: This overrides the base delete method to add documentation
        # but uses the parent's implementation which now handles async responses
        super().delete(wait=wait)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from pylxd import managers
from pylxd.models import _model as model


class StoragePool(model.Model):
    """An LXD storage_pool.

    This corresponds to the LXD endpoint at
    /1.0/storage-pools

    api_extension: 'storage'
    """

    name = model.Attribute(readonly=True)
    driver = model.Attribute(readonly=True)
    used_by = model.Attribute(readonly=True)
    config = model.Attribute()
    managed = model.Attribute(readonly=True)
    description = model.Attribute()
    status = model.Attribute(readonly=True)
    locations = model.Attribute(readonly=True)

    resources = model.Manager()
    volumes = model.Manager()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.resources = managers.StorageResourcesManager(self)
        self.volumes = managers.StorageVolumeManager(self)

    @classmethod
    def get(cls, client, name):
        """Get a storage_pool by name.

        Implements GET /1.0/storage-pools/<name>

        :param client: The pylxd client object
        :type client: :class:`pylxd.client.Client`
        :param name: the name of the storage pool to get
        :type name: str
        :returns: a storage pool if successful, raises NotFound if not found
        :rtype: :class:`pylxd.models.storage_pool.StoragePool`
        :raises: :class:`pylxd.exceptions.NotFound`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        """
        client.assert_has_api_extension("storage")
        response = client.api.storage_pools[name].get()

        storage_pool = cls(client, **response.json()["metadata"])
        return storage_pool

    @classmethod
    def all(cls, client):
        """Get all storage_pools.

        Implements GET /1.0/storage-pools

        Note that the returned list is 'sparse' in that only the name of the
        pool is populated.  If any of the attributes are used, then the `sync`
        function is called to populate the object fully.

        :param client: The pylxd client object
        :type client: :class:`pylxd.client.Client`
        :returns: a storage pool if successful, raises NotFound if not found
        :rtype: [:class:`pylxd.models.storage_pool.StoragePool`]
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        """
        client.assert_has_api_extension("storage")
        response = client.api.storage_pools.get()

        storage_pools = []
        for url in response.json()["metadata"]:
            name = url.split("/")[-1]
            storage_pools.append(cls(client, name=name))
        return storage_pools

    @classmethod
    def create(cls, client, definition, wait=True):
        """Create a storage_pool from config.

        Implements POST /1.0/storage-pools

        The `definition` parameter defines what the storage pool will be.
        An example config for the zfs driver is::

            {
                "config": {
                    "size": "10GB"
                },
                "driver": "zfs",
                "name": "pool1"
            }

        Note that **all** fields in the `definition` parameter are strings.

        For further details on the storage pool types see:
        https://documentation.ubuntu.com/lxd/en/latest/explanation/storage/

        The function returns the a `StoragePool` instance, if it is
        successfully created, otherwise an Exception is raised.

        :param client: The pylxd client object
        :type client: :class:`pylxd.client.Client`
        :param definition: the fields to pass to the LXD API endpoint
        :type definition: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :returns: a storage pool if successful, raises NotFound if not found
        :rtype: :class:`pylxd.models.storage_pool.StoragePool`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            couldn't be created.
        """
        client.assert_has_api_extension("storage")
        response = client.api.storage_pools.post(json=definition)

        # Use helper method for async handling
        cls._handle_async_response_for_client(client, response, wait)
        storage_pool = cls.get(client, definition["name"])
        return storage_pool

    @classmethod
    def exists(cls, client, name):
        """Determine whether a storage pool exists.

        A convenience method to determine a pool exists.  However, it is better
        to try to fetch it and catch the :class:`pylxd.exceptions.NotFound`
        exception, as otherwise the calling code is like to fetch the pool
        twice.  Only use this if the calling code doesn't *need* the actual
        storage pool information.

        :param client: The pylxd client object
        :type client: :class:`pylxd.client.Client`
        :param name: the name of the storage pool to get
        :type name: str
        :returns: True if the storage pool exists, False if it doesn't.
        :rtype: bool
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        """
        try:
            cls.get(client, name)
            return True
        except cls.NotFound:
            return False

    @property
    def api(self):
        """Provides an object with the endpoint:

        /1.0/storage-pools/<self.name>

        Used internally to construct endpoints.

        :returns: an API node with the named endpoint
        :rtype: :class:`pylxd.client._APINode`
        """
        return self.client.api.storage_pools[self.name]

    def save(self, wait=False):
        """Save the model using PUT back to the LXD server.

        Implements PUT /1.0/storage-pools/<self.name> *automagically*

        The fields affected are: `description` and `config`.  Note that they
        are replaced in their *entirety*.  If finer grained control is
        required, please use the
        :meth:`~pylxd.models.storage_pool.StoragePool.patch` method directly.

        Updating a storage pool may fail if the config is not acceptable to
        LXD. An :class:`~pylxd.exceptions.LXDAPIException` will be generated in
        that case.

        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be deleted.
        """
        # Note this method exists so that it is documented via sphinx.
        super().save(wait=wait)

    def delete(self, wait=False):
        """Delete the storage pool.

        Implements DELETE /1.0/storage-pools/<self.name>

        Deleting a storage pool may fail if it is being used.  See the LXD
        documentation for further details.

        :param wait: Whether to wait for async operations to complete.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be deleted.
        """
        # Note this method exists so that it is documented via sphinx.
        super().delete(wait=wait)

    def put(self, put_object, wait=False):
        """Put the storage pool.

        Implements PUT /1.0/storage-pools/<self.name>

        Putting to a storage pool may fail if the new configuration is
        incompatible with the pool.  See the LXD documentation for further
        details.

        Note that the object is refreshed with a `sync` if the PUT is
        successful.  If this is *not* desired, then the raw API on the client
        should be used.

        :param put_object: A dictionary.  The most useful key will be the
            `config` key.
        :type put_object: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be modified.
        """
        # Note this method exists so that it is documented via sphinx.
        super().put(put_object, wait)

    def patch(self, patch_object, wait=False):
        """Patch the storage pool.

        Implements PATCH /1.0/storage-pools/<self.name>

        Patching the object allows for more fine grained changes to the config.
        The object is refreshed if the PATCH is successful.  If this is *not*
        required, then use the client api directly.

        :param patch_object: A dictionary.  The most useful key will be the
            `config` key.
        :type patch_object: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be modified.
        """
        # Note this method exists so that it is documented via sphinx.
        super().patch(patch_object, wait)


class StorageResources(model.Model):
    """An LXD Storage Resources model.

    This corresponds to the LXD endpoing at
    /1.0/storage-pools/<pool>/resources

    At present, this is read-only model.

    api_extension: 'resources'
    """

    space = model.Attribute(readonly=True)
    inodes = model.Attribute(readonly=True)

    @classmethod
    def get(cls, storage_pool):
        """Get a storage_pool resource for a named pool

        Implements GET /1.0/storage-pools/<pool>/resources

        Needs the 'resources' api extension in the LXD server.

        :param storage_pool: a storage pool object on which to fetch resources
        :type storage_pool: :class:`pylxd.models.storage_pool.StoragePool`
        :returns: A storage resources object
        :rtype: :class:`pylxd.models.storage_pool.StorageResources`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'resources' api extension is missing.
        """
        storage_pool.client.assert_has_api_extension("resources")
        response = storage_pool.api.resources.get()
        resources = cls(storage_pool.client, **response.json()["metadata"])
        return resources


class StorageVolume(model.Model):
    """An LXD Storage volume.

    This corresponds to the LXD endpoing at
    /1.0/storage-pools/<pool>/volumes

    api_extension: 'storage'
    """

    name = model.Attribute(readonly=True)
    type = model.Attribute(readonly=True)
    content_type = model.Attribute(readonly=True)
    description = model.Attribute(readonly=True)
    config = model.Attribute()
    used_by = model.Attribute(readonly=True)
    location = model.Attribute(readonly=True)
    # Date strings follow the ISO 8601 pattern
    created_at = model.Attribute(readonly=True)
    pool = model.Attribute(readonly=True)
    project = model.Attribute(readonly=True)

    snapshots = model.Manager()

    storage_pool = model.Parent()

    @property
    def api(self):
        """Provides an object with the endpoint:

        /1.0/storage-pools/<storage_pool.name>/volumes/<self.type>/<self.name>

        Used internally to construct endpoints.

        :returns: an API node with the named endpoint
        :rtype: :class:`pylxd.client._APINode`
        """
        return self.storage_pool.api.volumes[self.type][self.name]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.snapshots = managers.StorageVolumeSnapshotManager(self)

    @classmethod
    def all(cls, storage_pool):
        """Get all the volumnes for this storage pool.

        Implements GET /1.0/storage-pools/<name>/volumes

        Volumes returned from this method will only have the name
        set, as that is the only property returned from LXD. If more
        information is needed, `StorageVolume.sync` is the method call
        that should be used.

        Note that the storage volume types are 'container', 'image' and
        'custom', and these maps to the names 'containers', 'images' and
        everything else is mapped to 'custom'.

        :param storage_pool: a storage pool object on which to fetch resources
        :type storage_pool: :class:`pylxd.models.storage_pool.StoragePool`
        :returns: a list storage volume if successful
        :rtype: [:class:`pylxd.models.storage_pool.StorageVolume`]
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        """
        storage_pool.client.assert_has_api_extension("storage")
        response = storage_pool.api.volumes.get()

        volumes = []
        for volume in response.json()["metadata"]:
            (_type, name) = volume.split("/")[-2:]
            # for each type, convert to the string that will work with GET
            if _type == "container":
                _type = "container"
            elif _type == "virtual-machine":
                _type = "virtual-machine"
            elif _type == "instance":
                _type = "instance"
            elif _type == "image":
                _type = "image"
            else:
                _type = "custom"
            volumes.append(
                cls(
                    storage_pool.client,
                    name=name,
                    type=_type,
                    storage_pool=storage_pool,
                )
            )
        return volumes

    @classmethod
    def get(cls, storage_pool, _type, name):
        """Get a StorageVolume by type and name.

        Implements GET /1.0/storage-pools/<pool>/volumes/<type>/<name>

        The `_type` param can only (currently) be one of 'container', 'image'
        or 'custom'.  This was determined by read the LXD source.

        :param storage_pool: a storage pool object on which to fetch resources
        :type storage_pool: :class:`pylxd.models.storage_pool.StoragePool`
        :param _type: the type; one of 'container', 'image', 'custom'
        :type _type: str
        :param name: the name of the storage volume to get
        :type name: str
        :returns: a storage pool if successful, raises NotFound if not found
        :rtype: :class:`pylxd.models.storage_pool.StorageVolume`
        :raises: :class:`pylxd.exceptions.NotFound`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        """
        storage_pool.client.assert_has_api_extension("storage")
        response = storage_pool.api.volumes[_type][name].get()

        volume = cls(
            storage_pool.client,
            storage_pool=storage_pool,
            **response.json()["metadata"],
        )
        return volume

    @classmethod
    # def create(cls, storage_pool, definition, wait=True, *args):
    def create(cls, storage_pool, *args, **kwargs):
        """Create a 'custom' Storage Volume in the associated storage pool.

        Implements POST /1.0/storage-pools/<pool>/volumes/custom

        See https://documentation.ubuntu.com/lxd/en/latest/api/#/storage/storage_pool_volumes_type_post
        for more details on what the `definition` parameter dictionary should
        contain for various volume creation.

        At the moment the only type of volume that can be created is 'custom',
        and this is currently hardwired into the function.

        The function signature 'hides' that the first parameter to the function
        is the definition.  The function should be called as:

        >> a_storage_pool.volumes.create(definition_dict, wait=<bool>)

        where `definition_dict` is mandatory, and `wait` defaults to True,
        which makes the default a synchronous function call.

        Note that **all** fields in the `definition` parameter are strings.

        If the caller doesn't wan't to wait for an async operation, then it
        MUST be passed as a keyword argument, and not as a positional
        substitute.

        The function returns the a
        :class:`~pylxd.models.storage_pool.StoragePool` instance, if it is
        successfully created, otherwise an Exception is raised.

        :param storage_pool: a storage pool object on which to fetch resources
        :type storage_pool: :class:`pylxd.models.storage_pool.StoragePool`
        :param definition: the fields to pass to the LXD API endpoint
        :type definition: dict
        :param wait: wait until an async action has completed (default True)
        :type wait: bool
        :returns: a storage pool volume if successful, raises NotFound if not
            found
        :rtype: :class:`pylxd.models.storage_pool.StorageVolume`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            volume couldn't be created.
        """
        # This is really awkward, but the implementation details mean,
        # depending on how this function is called, we can't know whether the
        # 2nd positional argument will be definition or a client object. This
        # is an difficulty with how BaseManager is implemented, and to get the
        # convenience of being able to call this 'naturally' off of a
        # storage_pool.  So we have to jump through some hurdles to get the
        # right positional parameters.
        storage_pool.client.assert_has_api_extension("storage")
        wait = kwargs.get("wait", True)
        definition = args[-1]
        assert isinstance(definition, dict)
        assert "name" in definition
        response = storage_pool.api.volumes.custom.post(json=definition)

        # Use class method helper for async handling
        cls._handle_async_response_for_client(storage_pool.client, response, wait)

        volume = cls.get(storage_pool, "custom", definition["name"])
        return volume

    def rename(self, _input, wait=False):
        """Rename a storage volume

        This requires api_extension: 'storage_api_volume_rename'.

        Implements: POST /1.0/storage-pools/<pool>/volumes/<type>/<name>

        This operation is either sync or async (when moving to a different
        pool).

        This command also allows the migration across instances, and therefore
        it returns the metadata section (as a dictionary) from the call.  The
        caller should assume that the object is stale and refetch it after any
        migrations are done.  However, the 'name' attribute of the object is
        updated for consistency.

        Unlike :meth:`~pylxd.models.storage_pool.StorageVolume.create`, this
        method does not override any items in the input definition, although it
        does check that the 'name' and 'pool' parameters are set.

        Please see: https://documentation.ubuntu.com/lxd/en/latest/api/#/storage/storage_pool_volume_type_post
        for more details.

        :param _input: The `input` specification for the rename.
        :type _input: dict
        :param wait: Wait for an async operation, if it is async.
        :type wait: bool
        :returns:  The dictionary from the metadata section of the response if
            successful.
        :rtype: dict[str, str]
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage_api_volume_rename' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            volume couldn't be renamed.
        """
        assert isinstance(_input, dict)
        assert "name" in _input
        assert "pool" in _input
        response = self.api.post(json=_input)

        # Use instance method helper for async handling
        self._handle_async_response(response, wait)

        self.name = _input["name"]
        return response.json()["metadata"]

    def put(self, put_object, wait=False):
        """Put the storage volume.

        Implements: PUT /1.0/storage-pools/<pool>/volumes/<type>/<name>

        Note that this is functionality equivalent to
        :meth:`~pyxld.models.storage_pool.StorageVolume.save` but by using a
        new object (`put_object`) rather than modifying the object and then
        calling :meth:`~pyxld.models.storage_pool.StorageVolume.save`.

        Putting to a storage volume may fail if the new configuration is
        incompatible with the pool.  See the LXD documentation for further
        details.

        Note that the object is refreshed with a `sync` if the PUT is
        successful.  If this is *not* desired, then the raw API on the client
        should be used.

        :param put_object: A dictionary.  The most useful key will be the
            `config` key.
        :type put_object: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be modified.
        """
        # Note this method exists so that it is documented via sphinx.
        super().put(put_object, wait)

    def patch(self, patch_object, wait=False):
        """Patch the storage volume.

        Implements: PATCH /1.0/storage-pools/<pool>/volumes/<type>/<name>

        Patching the object allows for more fine grained changes to the config.
        The object is refreshed if the PATCH is successful.  If this is *not*
        required, then use the client api directly.

        :param patch_object: A dictionary.  The most useful key will be the
            `config` key.
        :type patch_object: dict
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage
            volume can't be modified.
        """
        # Note this method exists so that it is documented via sphinx.
        super().patch(patch_object, wait)

    def save(self, wait=False):
        """Save the model using PUT back to the LXD server.

        Implements: PUT /1.0/storage-pools/<pool>/volumes/<type>/<name>
        *automagically*.

        The field affected is `config`.  Note that it is replaced *entirety*.
        If finer grained control is required, please use the
        :meth:`~pylxd.models.storage_pool.StorageVolume.patch` method directly.

        Updating a storage volume may fail if the config is not acceptable to
        LXD. An :class:`~pylxd.exceptions.LXDAPIException` will be generated in
        that case.

        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage
            volume can't be saved.
        """
        # Note this method exists so that it is documented via sphinx.
        super().save(wait=wait)

    def delete(self, wait=False):
        """Delete the storage pool.

        Implements: DELETE /1.0/storage-pools/<pool>/volumes/<type>/<name>

        Deleting a storage volume may fail if it is being used.
        See https://documentation.ubuntu.com/lxd/en/latest/explanation/storage/#storage-volumes
        for further details.

        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be deleted.
        """
        # Note this method exists so that it is documented via sphinx.
        super().delete(wait=wait)

    def restore_from(self, snapshot_name, wait=False):
        """Restore this volume from a snapshot using its name.

        Attempts to restore a volume using a snapshot identified by its name.

        Implements POST /1.0/storage-pools/<pool>/volumes/custom/<volume_name>/snapshot/<name>

        :param snapshot_name: the name of the snapshot to restore from
        :type snapshot_name: str
        :param wait: wait until the operation is completed.
        :type wait: boolean
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' or 'storage_api_volume_snapshots' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the the operation fails.
        :returns: the original response from the restore operation (not the
            operation result)
        :rtype: :class:`requests.Response`
        """
        response = self.api.put(json={"restore": snapshot_name})

        # Handle both sync and async responses for endpoint changing from sync to async
        if wait:
            response_json = response.json()
            if response_json["type"] == "async":
                self.client.operations.wait_for_operation(response_json["operation"])
        return response


class StorageVolumeSnapshot(model.Model):
    """A storage volume snapshot.

    This corresponds to the LXD endpoing at
    /1.0/storage-pools/<pool>/volumes/<type>/<volume>/snapshots

    api_extension: 'storage_api_volume_snapshots'
    """

    name = model.Attribute(readonly=True)
    description = model.Attribute()
    config = model.Attribute()
    content_type = model.Attribute(readonly=True)
    # Date strings follow the ISO 8601 pattern
    created_at = model.Attribute(readonly=True)
    expires_at = model.Attribute()

    _endpoint = "snapshots"

    volume = model.Parent()

    @property
    def api(self):
        """Provides an object with the endpoint:

        /1.0/storage-pools/<volume.storage_pool.name>/volumes/<volume.type>/<volume.name>/snapshots/<self.name>

        Used internally to construct endpoints.

        :returns: an API node with the named endpoint
        :rtype: :class:`pylxd.client._APINode`
        """
        return self.volume.api[self._endpoint][self.name]

    @classmethod
    def __parse_snapshot_json(cls, volume, snapshot_json):
        snapshot_object = cls(volume.client, volume=volume, **snapshot_json)

        # Snapshot names are namespaced in LXD, as volume-name/snapshot-name.
        # We hide that implementation detail.
        snapshot_object.name = snapshot_object.name.split("/")[-1]

        # If response does not include expires_at sync the object to get that attribute.
        if not snapshot_json.get("expires_at"):
            snapshot_object.sync()

        # Getting '0001-01-01T00:00:00Z' means that the volume does not have an expiration time set.
        if snapshot_object.expires_at == "0001-01-01T00:00:00Z":
            snapshot_object.expires_at = None

        # Overriding default value for when created_at is not present on the response due to using LXD 4.0.
        # Also having created_at as "0001-01-01T00:00:00Z" means this information is not available.
        # This could be because the information was lost or the snapshot was created using an older LXD version.
        if (
            not volume.client.has_api_extension("storage_volumes_created_at")
            or snapshot_json["created_at"] == "0001-01-01T00:00:00Z"
        ):
            snapshot_object.created_at = None

        # This field is may empty so derive it from its volume.
        if not snapshot_object.content_type:
            snapshot_object.content_type = volume.content_type

        return snapshot_object

    @classmethod
    def get(cls, volume, name):
        """Get a :class:`pylxd.models.StorageVolumeSnapshot` by its name.

        Implements GET /1.0/storage-pools/<pool>/volumes/custom/<volume_name>/snapshots/<name>

        :param client: a storage pool object on which to fetch resources
        :type storage_pool: :class:`pylxd.models.storage_pool.StoragePool`
        :param _type: the volume type; one of 'container', 'image', 'custom'
        :type _type: str
        :param volume: the name of the storage volume snapshot to get
        :type volume: pylxd.models.StorageVolume
        :returns: a storage pool if successful, raises NotFound if not found
        :rtype: :class:`pylxd.models.storage_pool.StorageVolume`
        :raises: :class:`pylxd.exceptions.NotFound`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' or 'storage_api_volume_snapshots' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the the operation fails.
        """
        volume.client.assert_has_api_extension("storage_api_volume_snapshots")

        response = volume.api.snapshots[name].get()

        return cls.__parse_snapshot_json(volume, response.json()["metadata"])

    @classmethod
    def all(cls, volume, use_recursion=False):
        """Get all :class:`pylxd.models.StorageVolumeSnapshot` objects related to a certain volume.
        If use_recursion is unset or set to False, a list of snapshot names is returned.
        If use_recursion is set to True, a list of :class:`pylxd.models.StorageVolumeSnapshot` objects is returned
        containing additional information for each snapshot.

        Implements GET /1.0/storage-pools/<pool>/volumes/custom/<volume_name>/snapshots

        :param volume: The storage volume snapshot to get snapshots from
        :type volume: pylxd.models.StorageVolume
        :param use_recursion: Specifies whether 'recursion=1' should be used on the request.
        :type use_recursion: bool
        :returns: A list of storage volume snapshot names if use_recursion is False, otherwise
            returns a list of :class:`pylxd.models.StorageVolumeSnapshot`
        :rtype: list
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage' or 'storage_api_volume_snapshots' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the the operation fails.
        """
        volume.client.assert_has_api_extension("storage_api_volume_snapshots")

        if use_recursion:
            # Using recursion so returning list of StorageVolumeSnapshot objects.
            response = volume.api.snapshots.get(params={"recursion": 1})

            return [
                cls.__parse_snapshot_json(volume, snapshot)
                for snapshot in response.json()["metadata"]
            ]

        response = volume.api.snapshots.get()

        return [
            snapshot_name.split("/")[-1]
            for snapshot_name in response.json()["metadata"]
        ]

    @classmethod
    def create(cls, volume, name=None, expires_at=None, wait=True):
        """Create new :class:`pylxd.models.StorageVolumeSnapshot` object from the current volume state using the given attributes.

        Implements POST /1.0/storage-pools/<pool>/volumes/custom/<volume_name>/snapshots

        :param volume: :class:`pylxd.models.StorageVolume` object that represents the target volume to take the snapshot from
        :type volume: :class:`pylxd.models.StorageVolume`
        :param name: Optional parameter. Name of the created snapshot. The snapshot will be called "snap{index}" by default.
        :type name: str
        :param expires_at: Optional parameter. Expiration time for the created snapshot in ISO 8601 format. No expiration date by default.
        :type expires_at: str
        :param wait: Whether to wait for async operations to complete.
        :type wait: bool
        :returns: a storage volume snapshot if successful, raises an exception otherwise.
        :rtype: :class:`pylxd.models.StorageVolume`
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable` if the
            'storage_api_volume_snapshots' api extension is missing.
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the the operation fails.
        """
        volume.client.assert_has_api_extension("storage_api_volume_snapshots")

        response = volume.api.snapshots.post(
            json={"name": name, "expires_at": expires_at}
        )

        operation = None

        # Only parse JSON if we need to wait for async responses
        if wait:
            response_json = response.json()

            # Handle both sync and async responses
            if response_json["type"] == "async":
                operation = volume.client.operations.wait_for_operation(
                    response_json["operation"]
                )
            else:
                # Return the snapshot immediately without waiting for completion
                return volume.snapshots.get(name)

        # Extract the snapshot name from the response JSON in case it was not provided
        if not name:
            if operation and "storage_volume_snapshots" in operation.resources:
                name = operation.resources["storage_volume_snapshots"][0].split("/")[-1]
            else:
                # If using LXD 4.0, the snapshot name isn't provided on the request response
                # so grab the latest snapshot name instead.
                name = volume.snapshots.all()[-1].split("/")[-1]

        snapshot = volume.snapshots.get(name)
        return snapshot

    @classmethod
    def exists(cls, volume, name):
        """Determine whether a volume snapshot exists in LXD.

        :param name: Name of the desired snapshot.
        :type name: str
        :returns: True if a snapshot with the given name exists, returns False otherwise.
        :rtype: bool
        :raises: `pylxd.exceptions.LXDAPIException` if the the operation fails.
        """
        try:
            volume.snapshots.get(name)
            return True
        except cls.NotFound:
            return False

    def rename(self, new_name, wait=True):
        """Rename a storage volume snapshot.

        Implements POST /1.0/storage-pools/<pool>/volumes/custom/<volume_name>/snapshot/<name>

        Renames a storage volume snapshot, changing the endpoints that reference it.

        :param new_name: The new name for the snapshot
        :type new_name: str
        :param wait: Whether to wait for async operations to complete
        :type wait: bool
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the the operation fails.
        """
        response = self.api.post(json={"name": new_name})

        # Use instance method helper for async handling
        self._handle_async_response(response, wait)

        self.name = new_name

    def restore(self, wait=False):
        """Restore the volume from this snapshot.

        Attempts to restore a custom volume using this snapshot.
        Equivalent to pylxd.models.StorageVolume.restore_from(this_snapshot).

        :param wait: wait until the operation is completed.
        :type wait: boolean
        :raises: :class:`pylxd.exceptions.LXDAPIException` if the the operation fails.
        """
        self.volume.restore_from(self.name, wait)

    def delete(self, wait=False):
        """Delete this storage pool snapshot.

        Implements: DELETE /1.0/storage-pools/<pool>/volumes/custom/<volume_name>/snapshot/<name>

        :raises: :class:`pylxd.exceptions.LXDAPIException` if the storage pool
            can't be deleted.
        """
        super().delete(wait=wait)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

from pylxd import managers
from pylxd.models import _model as model


class NetworkForward(model.Model):
    config = model.Attribute()
    description = model.Attribute()
    location = model.Attribute()
    listen_address = model.Attribute()
    ports = model.Attribute()

    network = model.Parent()

    @classmethod
    def get(cls, client, network, listen_address):
        response = client.api.networks[network.name].forwards[listen_address].get()
        forward = cls(client, network=network, **response.json()["metadata"])
        return forward

    @classmethod
    def create(cls, client, network, config):
        client.api.networks[network.name].forwards.post(json=config)

        return cls(client, network=network, **config)

    def save(self, *args, **kwargs):
        self.client.assert_has_api_extension("network")
        super().save(*args, **kwargs)

    @property
    def api(self):
        return self.client.api.networks[self.network.name].forwards[self.listen_address]

    def __str__(self):
        return json.dumps(self.marshall(skip_readonly=False), indent=2)

    def __repr__(self):
        attrs = []
        for attribute, value in self.marshall().items():
            attrs.append(f"{attribute}={json.dumps(value, sort_keys=True)}")

        return f"{self.__class__.__name__}({', '.join(sorted(attrs))})"


class NetworkState(model.AttributeDict):
    """A simple object for representing a network state."""


class Network(model.Model):
    """Model representing a LXD network."""

    name = model.Attribute()
    description = model.Attribute()
    type = model.Attribute()
    config = model.Attribute()
    status = model.Attribute(readonly=True)
    locations = model.Attribute(readonly=True)
    managed = model.Attribute(readonly=True)
    used_by = model.Attribute(readonly=True)
    _endpoint = "networks"

    forwards = model.Manager()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.forwards = managers.NetworkForwardManager(self.client, self)

    @classmethod
    def exists(cls, client, name):
        """
        Determine whether network with provided name exists.

        :param client: client instance
        :type client: :class:`~pylxd.client.Client`
        :param name: name of the network
        :type name: str
        :returns: `True` if network exists, `False` otherwise
        :rtype: bool
        """
        try:
            client.networks.get(name)
            return True
        except cls.NotFound:
            return False

    @classmethod
    def get(cls, client, name):
        """
        Get a network by name.

        :param client: client instance
        :type client: :class:`~pylxd.client.Client`
        :param name: name of the network
        :type name: str
        :returns: network instance (if exists)
        :rtype: :class:`Network`
        :raises: :class:`~pylxd.exceptions.NotFound` if network does not exist
        """
        response = client.api.networks[name].get()

        return cls(client, **response.json()["metadata"])

    @classmethod
    def all(cls, client):
        """
        Get all networks.

        :param client: client instance
        :type client: :class:`~pylxd.client.Client`
        :rtype: list[:class:`Network`]
        """
        response = client.api.networks.get()

        networks = []
        for url in response.json()["metadata"]:
            name = url.split("/")[-1]
            networks.append(cls(client, name=name))
        return networks

    @classmethod
    def create(cls, client, name, description=None, type=None, config=None):
        """
        Create a network.

        :param client: client instance
        :type client: :class:`~pylxd.client.Client`
        :param name: name of the network
        :type name: str
        :param description: description of the network
        :type description: str
        :param type: type of the network
        :type type: str
        :param config: additional configuration
        :type config: dict
        """
        client.assert_has_api_extension("network")

        network = {"name": name}
        if description is not None:
            network["description"] = description
        if type is not None:
            network["type"] = type
        if config is not None:
            network["config"] = config
        client.api.networks.post(json=network)
        return cls.get(client, name)

    def rename(self, new_name):
        """
        Rename a network.

        :param new_name: new name of the network
        :type new_name: str
        :return: Renamed network instance
        :rtype: :class:`Network`
        """
        self.client.assert_has_api_extension("network")
        self.client.api.networks.post(json={"name": new_name})
        return Network.get(self.client, new_name)

    def save(self, *args, **kwargs):
        self.client.assert_has_api_extension("network")
        super().save(*args, **kwargs)

    def state(self):
        """Get network state."""
        response = self.api.state.get()
        state = NetworkState(response.json()["metadata"])
        return state

    @property
    def api(self):
        return self.client.api.networks[self.name]

    def __str__(self):
        return json.dumps(self.marshall(skip_readonly=False), indent=2)

    def __repr__(self):
        attrs = []
        for attribute, value in self.marshall().items():
            attrs.append(f"{attribute}={json.dumps(value, sort_keys=True)}")

        return f"{self.__class__.__name__}({', '.join(sorted(attrs))})"



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from integration.testing import IntegrationTestCase


class ClusterMemberTestCase(IntegrationTestCase):
    def setUp(self):
        super().setUp()

        if not self.client.has_api_extension("clustering"):
            self.skipTest("Required LXD API extension not available!")

        if not self.client.server_clustered:
            self.skipTest("Server isn't part of a cluster!")


class TestClusterMembers(ClusterMemberTestCase):
    """Tests for `Client.cluster_members.`"""

    def test_get(self):
        """A cluster member is fetched by its name."""

        members = self.client.cluster.members.all()

        random_member_name = f"{members[0].server_name}"
        random_member_url = f"{members[0].url}"

        member = self.client.cluster.members.get(random_member_name)

        new_url = f"{member.url}"
        self.assertEqual(random_member_url, new_url)



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from integration.testing import IntegrationTestCase
from pylxd import exceptions


class TestProfiles(IntegrationTestCase):
    """Tests for `Client.profiles.`"""

    def test_get(self):
        """A profile is fetched by name."""
        name = self.create_profile()
        self.addCleanup(self.delete_profile, name)

        profile = self.client.profiles.get(name)

        self.assertEqual(name, profile.name)

    def test_all(self):
        """All profiles are fetched."""
        name = self.create_profile()
        self.addCleanup(self.delete_profile, name)

        profiles = self.client.profiles.all()

        self.assertIn(name, [profile.name for profile in profiles])

    def test_create(self):
        """A profile is created."""
        name = "an-profile"
        config = {"limits.memory": "1GB"}
        description = "Memory limiting profile"
        profile = self.client.profiles.create(name, config, description=description)
        self.addCleanup(self.delete_profile, name)

        self.assertEqual(name, profile.name)
        self.assertEqual(config, profile.config)
        self.assertEqual(description, profile.description)


class TestProfile(IntegrationTestCase):
    """Tests for `Profile`."""

    def setUp(self):
        super().setUp()
        name = self.create_profile()
        self.profile = self.client.profiles.get(name)

    def tearDown(self):
        super().tearDown()
        self.delete_profile(self.profile.name)

    def test_save(self):
        """A profile is updated."""
        self.profile.config["limits.memory"] = "16GB"
        self.profile.save(wait=True)

        profile = self.client.profiles.get(self.profile.name)
        self.assertEqual("16GB", profile.config["limits.memory"])

    def test_rename(self):
        """A profile is renamed."""
        name = "a-other-profile"
        self.addCleanup(self.delete_profile, name)

        self.profile.rename(name, wait=True)
        profile = self.client.profiles.get(name)

        self.assertEqual(name, profile.name)

    def test_delete(self):
        """A profile is deleted."""
        self.profile.delete(wait=True)

        self.assertRaises(
            exceptions.LXDAPIException, self.client.profiles.get, self.profile.name
        )



# Copyright (c) 2016 Canonical Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import base64
import json
import os
import re
import socket
from enum import Enum
from typing import NamedTuple
from urllib import parse

import requests
import requests.adapters
import urllib3
import urllib3.connection
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from ws4py.client import WebSocketBaseClient

from pylxd import exceptions, managers

SNAP_ROOT = os.path.expanduser("~/snap/lxd/common/config/")
APT_ROOT = os.path.expanduser("~/.config/lxc/")
CERT_FILE_NAME = "client.crt"
KEY_FILE_NAME = "client.key"
# check that the cert file and key file exist at the appropriate path
if os.path.exists(os.path.join(SNAP_ROOT, CERT_FILE_NAME)):  # pragma: no cover
    CERTS_PATH = SNAP_ROOT  # pragma: no cover
else:  # pragma: no cover
    CERTS_PATH = APT_ROOT  # pragma: no cover


class Cert(NamedTuple):
    """Paths for a certificate/key pair."""

    cert: str
    key: str


DEFAULT_CERTS = Cert(
    cert=os.path.join(CERTS_PATH, CERT_FILE_NAME),
    key=os.path.join(CERTS_PATH, KEY_FILE_NAME),
)  # pragma: no cover

DEFAULT_SCHEME = "http+unix://"
SOCKET_CONNECTION_TIMEOUT = 60


class EventType(Enum):
    All = "all"
    Operation = "operation"
    Logging = "logging"
    Lifecycle = "lifecycle"


class _UnixSocketHTTPConnection(urllib3.connection.HTTPConnection):
    def __init__(self, unix_socket_url):
        super().__init__("localhost", timeout=SOCKET_CONNECTION_TIMEOUT)
        self.unix_socket_url = unix_socket_url
        self.timeout = SOCKET_CONNECTION_TIMEOUT
        self.sock = None

    def __del__(self):
        if self.sock:
            self.sock.close()

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        socket_path = parse.unquote(parse.urlparse(self.unix_socket_url).netloc)
        sock.connect(socket_path)
        self.sock = sock


class _UnixSocketHTTPConnectionPool(urllib3.HTTPConnectionPool):
    def __init__(self, socket_path):
        super().__init__("localhost")
        self.socket_path = socket_path

    def _new_conn(self):
        return _UnixSocketHTTPConnection(self.socket_path)


class _UnixAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, pool_connections=25, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pools = urllib3._collections.RecentlyUsedContainer(
            pool_connections, dispose_func=lambda p: p.close()
        )

    def get_connection(self, url, proxies):
        with self.pools.lock:
            conn = self.pools.get(url)
            if conn:
                return conn

            conn = _UnixSocketHTTPConnectionPool(url)
            self.pools[url] = conn

        return conn

    # This method is needed fo compatibility with later requests versions.
    def get_connection_with_tls_context(self, request, verify, proxies=None, cert=None):
        return self.get_connection(request.url, None)

    def request_url(self, request, proxies):
        return request.path_url

    def close(self):
        self.pools.clear()


class LXDSSLAdapter(requests.adapters.HTTPAdapter):
    def cert_verify(self, conn, url, verify, cert):
        with open(verify, "rb") as fd:
            servercert = x509.load_pem_x509_certificate(fd.read())
            fingerprint = servercert.fingerprint(hashes.SHA256())

        conn.assert_fingerprint = "".join([f"{i:02x}" for i in fingerprint])
        super().cert_verify(conn, url, False, cert)


def get_session_for_url(url: str, verify=None, cert=None) -> requests.Session:
    """Create a Session for use with requests for the given URL.

    Call sites can use this to customise the session before passing into a Client.
    """
    session = requests.Session()
    if url.startswith(DEFAULT_SCHEME):
        session.mount(DEFAULT_SCHEME, _UnixAdapter())
    else:
        session.cert = cert
        session.verify = verify

        if isinstance(verify, str):
            session.mount(url, LXDSSLAdapter())
    return session


class _APINode:
    """An api node object.

    `session` is a `requests.Session` which is used for issuing
       requests. See `get_session_for_url`
    """

    def __init__(
        self,
        api_endpoint,
        session,
        timeout=None,
        project=None,
    ):
        self._api_endpoint = api_endpoint
        self._timeout = timeout
        self._project = project
        self.session = session

    def __getattr__(self, name):
        """Converts attribute lookup into the next /<segment> of an api
        url.

        :param name: the next segment
        :type name: str
        :returns: new _APINode with /<name> on the end
        :rtype: _APINode
        """
        # '-' can't be used in variable names
        if name in ("storage_pools", "virtual_machines"):
            name = name.replace("_", "-")
        return self.__class__(
            f"{self._api_endpoint}/{name}",
            session=self.session,
            timeout=self._timeout,
            project=self._project,
        )

    def __getitem__(self, item):
        """This converts python api.thing[name] -> ".../thing/name"

        :param item: the 'thing' in the square-braces in a python expr.
        :type item: str
        :returns: A new _APINode(with the new item tagged on as /<item>
        :rtype: _APINode
        """
        return self.__class__(
            f"{self._api_endpoint}/{item}",
            session=self.session,
            timeout=self._timeout,
            project=self._project,
        )

    def _assert_response(
        self, response, allowed_status_codes=(200,), stream=False, is_api=True
    ):
        """Assert properties of the response.

        LXD's API clearly defines specific responses. If the API
        response is something unexpected (i.e. an error), then
        we need to raise an exception and have the call points
        handle the errors or just let the issue be raised to the
        user.
        """
        if response.status_code not in allowed_status_codes:
            if response.status_code == 404:
                raise exceptions.NotFound(response)
            elif response.status_code == 409:
                raise exceptions.Conflict(response)
            raise exceptions.LXDAPIException(response)

        # In the case of streaming, we can't validate the json the way we
        # would with normal HTTP responses, so just ignore that entirely.
        # Likewize, we can ignore NON api calls as the contents don't need to
        # be validated.
        if stream or not is_api:
            return

        try:
            data = response.json()
        except ValueError:
            # Not a JSON response
            return

        if response.status_code == 200:
            # Synchronous request
            try:
                if data["type"] != "sync":
                    raise exceptions.LXDAPIException(response)
            except KeyError:
                # Missing 'type' in response
                raise exceptions.LXDAPIException(response)

    @property
    def scheme(self):
        return parse.urlparse(self.api._api_endpoint).scheme

    @property
    def netloc(self):
        return parse.urlparse(self.api._api_endpoint).netloc

    def get(self, *args, **kwargs):
        """Perform an HTTP GET.

        Note if 'is_api' is passed in the kwargs then it is popped and used to
        determine whether the get is an API call or a raw call.
        This is for py27 compatibility.
        """
        is_api = kwargs.pop("is_api", True)
        kwargs["timeout"] = kwargs.get("timeout", self._timeout)

        if self._project is not None:
            params = kwargs.get("params", {})
            params["project"] = self._project
            kwargs["params"] = params

        response = self.session.get(self._api_endpoint, *args, **kwargs)
        self._assert_response(
            response, stream=kwargs.get("stream", False), is_api=is_api
        )
        return response

    def post(self, *args, **kwargs):
        """Perform an HTTP POST."""
        kwargs["timeout"] = kwargs.get("timeout", self._timeout)

        target = kwargs.pop("target", None)
        if target is not None:
            params = kwargs.get("params", {})
            params["target"] = target
            kwargs["params"] = params

        if self._project is not None:
            params = kwargs.get("params", {})
            params["project"] = self._project
            kwargs["params"] = params

        response = self.session.post(self._api_endpoint, *args, **kwargs)
        # Prior to LXD 2.0.3, successful synchronous requests returned 200,
        # rather than 201.
        self._assert_response(response, allowed_status_codes=(200, 201, 202))
        return response

    def put(self, *args, **kwargs):
        """Perform an HTTP PUT."""
        kwargs["timeout"] = kwargs.get("timeout", self._timeout)

        if self._project is not None:
            params = kwargs.get("params", {})
            params["project"] = self._project
            kwargs["params"] = params

        response = self.session.put(self._api_endpoint, *args, **kwargs)
        self._assert_response(response, allowed_status_codes=(200, 202))
        return response

    def patch(self, *args, **kwargs):
        """Perform an HTTP PATCH."""
        kwargs["timeout"] = kwargs.get("timeout", self._timeout)

        if self._project is not None:
            params = kwargs.get("params", {})
            params["project"] = self._project
            kwargs["params"] = params

        response = self.session.patch(self._api_endpoint, *args, **kwargs)
        self._assert_response(response, allowed_status_codes=(200, 202))
        return response

    def delete(self, *args, **kwargs):
        """Perform an HTTP delete."""
        kwargs["timeout"] = kwargs.get("timeout", self._timeout)

        if self._project is not None:
            params = kwargs.get("params", {})
            params["project"] = self._project
            kwargs["params"] = params

        response = self.session.delete(self._api_endpoint, *args, **kwargs)
        self._assert_response(response, allowed_status_codes=(200, 202))
        return response


class _WebsocketClient(WebSocketBaseClient):
    """A basic websocket client for the LXD API.

    This client is intentionally barebones, and serves
    as a simple default. It simply connects and saves
    all json messages to a messages attribute, which can
    then be read are parsed.
    """

    def handshake_ok(self):
        self.messages = []

    def received_message(self, message):
        json_message = json.loads(message.data.decode("utf-8"))
        self.messages.append(json_message)


# Helper function used by Client.authenticate()
def _is_a_token(secret):
    """Inspect the provided secret to determine if it is a trust token.

    Try to base64 decode and parse the JSON to see if it contains a "secret" key.

    :param secret: The secret to inspect
    :type secret: str
    :returns: True if the secret is a trust token

    >>> _is_a_token("password")
    False
    >>> _is_a_token(base64.b64encode("password".encode("utf-8")))
    False
    >>> token = '{"client_name":"foo","fingerprint":"abcd","addresses":["192.0.2.1:8443"],"secret":"I-am-a-secret","expires_at":"0001-01-01T00:00:00Z","type":""}'
    >>> _is_a_token(base64.b64encode(json.dumps(token).encode("utf-8")))
    True
    """
    try:
        b64 = base64.b64decode(secret)
        token = json.loads(b64.decode("utf-8"))
        return "secret" in token
    except (TypeError, ValueError, json.JSONDecodeError, base64.binascii.Error):
        return False


class Client:
    """Client class for LXD REST API.

    This client wraps all the functionality required to interact with
    LXD, and is meant to be the sole entry point.

    .. attribute:: instances

        A :class:`models.Instance <pylxd.models.Instance>`.

    .. attribute:: containers

        A :class:`models.Container <pylxd.models.Container>`.

    .. attribute:: virtual_machines

        A :class:`models.VirtualMachine <pylxd.models.VirtualMachine>`.

    .. attribute:: images

        A :class:`models.Image <pylxd.models.Image>`.

    .. attribute:: operations

        A :class:`models.Operation <pylxd.models.Operation>`.

    .. attribute:: profiles

        A :class:`models.Profile <pylxd.models.Profile>`.

    .. attribute:: projects

        A :class:`models.Project <pylxd.models.Project>`.

    .. attribute:: api

        This attribute provides tree traversal syntax to LXD's REST API for
        lower-level interaction.

        Use the name of the url part as attribute or item of an api object to
        create another api object appended with the new url part name, ie:

            # /
            >> api = Client().api
            >> response = api.get()
            # Check status code and response
            >> print(response.status_code, response.json())
            # /instances/test/
            >> print(api.instances['test'].get().json())
    """

    def __init__(
        self,
        endpoint=None,
        version="1.0",
        cert=None,
        verify=True,
        timeout=None,
        project=None,
        session=None,
    ):
        """Constructs a LXD client

        :param endpoint: (optional): endpoint can be an http endpoint or
            a path to a unix socket.
        :param version: (optional): API version string to use with LXD
        :param cert: (optional): A tuple of (cert, key) to use with
            the http socket for client authentication
        :param verify: (optional): Either a boolean, in which case it controls
            whether we verify the server's TLS certificate, or a string, in
            which case it must be a path to a CA bundle to use.
            Defaults to ``True``.
        :param timeout: (optional) How long to wait for the server to send
            data before giving up, as a float, or a :ref:`(connect timeout,
            read timeout) <timeouts>` tuple.
        :param project: (optional) Name of the LXD project to interact with.
        :param session: (optional) A requests.Session to use for
            interactions with the endpoint. If given, cert and verify
            arguments are ignored - see get_session_for_url().
        """

        self.project = project
        if endpoint:
            if endpoint.startswith("/") and os.path.exists(endpoint):
                endpoint = f"http+unix://{parse.quote(endpoint, safe='')}"
            else:
                # Extra trailing slashes cause LXD to 301
                endpoint = endpoint.rstrip("/")
                if cert is None and (
                    os.path.exists(DEFAULT_CERTS.cert)
                    and os.path.exists(DEFAULT_CERTS.key)
                ):
                    cert = DEFAULT_CERTS

                # Try to use an existing server certificate if one exists
                if isinstance(verify, bool) and endpoint.startswith("https://"):
                    no_proto = re.sub(r"^https://\[?", "", endpoint)
                    remote = re.sub(r"]?(:[0-9]+)?$", "", no_proto)
                    remote_cert_path = os.path.join(
                        CERTS_PATH, "servercerts", remote + ".crt"
                    )
                    if os.path.exists(remote_cert_path):
                        verify = remote_cert_path
        else:
            if "LXD_DIR" in os.environ:
                path = os.path.join(os.environ.get("LXD_DIR"), "unix.socket")
            elif os.path.exists("/var/snap/lxd/common/lxd/unix.socket"):
                path = "/var/snap/lxd/common/lxd/unix.socket"
            else:
                path = "/var/lib/lxd/unix.socket"
            endpoint = f"http+unix://{parse.quote(path, safe='')}"
        self.cert = cert
        if session is None:
            session = get_session_for_url(endpoint, cert=cert, verify=verify)
        self.api = _APINode(
            f"{endpoint}/{version}", session, timeout=timeout, project=project
        )
        use_ssl = self.api.scheme == "https" and self.cert
        self.ssl_options = (
            {"certfile": self.cert[0], "keyfile": self.cert[1]} if use_ssl else None
        )

        # Verify the connection is valid.
        try:
            response = self.api.get()
            if response.status_code != 200:
                raise exceptions.ClientConnectionFailed()
            self.host_info = response.json()["metadata"]

        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.InvalidURL,
        ) as e:
            raise exceptions.ClientConnectionFailed(str(e))

        if self.project not in (None, "default") and not self.has_api_extension(
            "projects"
        ):
            raise exceptions.ClientConnectionFailed(
                "Remote server doesn't handle projects"
            )

        self.cluster = managers.ClusterManager(self)
        self.certificates = managers.CertificateManager(self)
        self.instances = managers.InstanceManager(self)
        self.containers = managers.ContainerManager(self)
        self.virtual_machines = managers.VirtualMachineManager(self)
        self.images = managers.ImageManager(self)
        self.networks = managers.NetworkManager(self)
        self.operations = managers.OperationManager(self)
        self.profiles = managers.ProfileManager(self)
        self.projects = managers.ProjectManager(self)
        self.storage_pools = managers.StoragePoolManager(self)
        self._resource_cache = None

    @property
    def trusted(self):
        return self.host_info["auth"] == "trusted"

    @property
    def server_clustered(self):
        return self.host_info["environment"].get("server_clustered", False)

    @property
    def resources(self):
        if self._resource_cache is None:
            self.assert_has_api_extension("resources")
            response = self.api.resources.get()
            if response.status_code != 200:
                raise exceptions.ClientConnectionFailed()
            self._resource_cache = response.json()["metadata"]
        return self._resource_cache

    def has_api_extension(self, name):
        """Return True if the `name` api extension exists.

        :param name: the api_extension to look for.
        :type name: str
        :returns: True if extension exists
        :rtype: bool
        """
        return name in self.host_info["api_extensions"]

    def assert_has_api_extension(self, name):
        """Asserts that the `name` api_extension exists.
        If not, then is raises the LXDAPIExtensionNotAvailable error.

        :param name: the api_extension to test for
        :type name: str
        :returns: None
        :raises: :class:`pylxd.exceptions.LXDAPIExtensionNotAvailable`
        """
        if not self.has_api_extension(name):
            raise exceptions.LXDAPIExtensionNotAvailable(name)

    def authenticate(self, secret, use_token_auth=True):
        if self.trusted:
            return

        if not isinstance(self.api.session.cert, tuple):
            raise exceptions.ClientConnectionFailed("No client certificate specified")

        try:
            with open(self.api.session.cert[0]) as f:
                cert = f.read().encode("utf-8")
        except FileNotFoundError:
            raise exceptions.ClientConnectionFailed("Client certificate not found")

        # Quirk to handle 5.21 that supports explicit trust tokens as well as
        # password auth. We need to ascertain if the provided secret is indeed a
        # token before trying to use it as such.
        secret_is_a_token = False
        if use_token_auth and self.has_api_extension("explicit_trust_token"):
            secret_is_a_token = _is_a_token(secret)

        if secret_is_a_token:
            self.certificates.create(password="", cert_data=cert, secret=secret)
        else:
            self.certificates.create(password=secret, cert_data=cert)

        # Refresh the host info
        response = self.api.get()
        self.host_info = response.json()["metadata"]

    @property
    def websocket_url(self):
        if self.api.scheme in ("http", "https"):
            host = self.api.netloc
            if self.api.scheme == "http":
                scheme = "ws"
            else:
                scheme = "wss"
        else:
            scheme = "ws+unix"
            host = parse.unquote(self.api.netloc)
        url = parse.urlunparse((scheme, host, "", "", "", ""))
        return url

    def events(self, websocket_client=None, event_types=None):
        """Get a websocket client for getting events.

        /events is a websocket url, and so must be handled differently than
        most other LXD API endpoints. This method returns
        a client that can be interacted with like any
        regular python socket.

        An optional `websocket_client` parameter can be
        specified for implementation-specific handling
        of events as they occur.

        :param websocket_client: Optional websocket client can be specified for
         implementation-specific handling of events as they occur.
        :type websocket_client: ws4py.client import WebSocketBaseClient

        :param event_types: Optional set of event types to propagate. Omit this
         argument or specify {EventTypes.All} to receive all events.
        :type event_types: Set[EventType]

        :returns: instance of the websocket client
        :rtype: Option[_WebsocketClient(), `websocket_client`]
        """
        if websocket_client is None:
            websocket_client = _WebsocketClient

        client = websocket_client(self.websocket_url, ssl_options=self.ssl_options)
        parsed = parse.urlparse(self.api.events._api_endpoint)

        resource = parsed.path

        if event_types and EventType.All not in event_types:
            query = parse.parse_qs(parsed.query)
            query.update({"type": ",".join(t.value for t in event_types)})
            resource = f"{resource}?{parse.urlencode(query)}"

        client.resource = resource

        return client
